package RtFC;
use strict;

BEGIN {
  use Exporter ();
  use vars qw( $VERSION @ISA @EXPORT @EXPORT_OK @EXPORT_FAIL );
  $VERSION = '0.8';
  @ISA = qw(Exporter);
  
  @EXPORT      = qw( OBSPORT UIPORT CORRPORT CORRINTERFACEPORT
		     RtFC_COMMAND_SIZE CORRCONFIGPATH_SIZE
		     GRAB_COMMAND ANTENNA_NAME_COMMAND
		     ANTENNA_ID_COMMAND ERROR_COMMAND REDEFINE_COMMAND
		     TRANSFER_COMMAND DATA_RECEIVED_COMMAND RTFC_MESSAGE
		     OBSCON_COMMAND CORRSETUP_COMMAND BASE_COMMAND
		     DUPLICATE_CONNECTION SHOWOBS_COMMAND
		     ENABLEOBS_COMMAND DISABLEOBS_COMMAND
		     DATA_TRANSFER 
		     CORR_CONNECTION OBSDATA
		     make_command
		     make_field make_commandfield make_simplecommandfield
		     recv_command);
  @EXPORT_OK   = qw ($RtFC_COMMAND_SIZE);
  @EXPORT_FAIL = qw ($imsg);

  use Carp;
}

use constant DEBUG => 0;

use constant OBSPORT => 7426;
use constant UIPORT => 7423;
use constant CORRPORT => 7420;
use constant CORRINTERFACEPORT => 8524;

use constant RtFC_COMMAND_SIZE => 256;
use constant CORRCONFIGPATH_SIZE => 256;

my $imsg = 0;
use constant BASE_COMMAND => $imsg++;
use constant GRAB_COMMAND => $imsg++;
use constant ANTENNA_NAME_COMMAND =>$imsg++;
use constant ANTENNA_ID_COMMAND => $imsg++;
use constant ERROR_COMMAND => $imsg++;
use constant REDEFINE_COMMAND => $imsg++;
use constant TRANSFER_COMMAND => $imsg++;
use constant DATA_RECEIVED_COMMAND => $imsg++;
use constant SHOWOBS_COMMAND => $imsg++;
use constant ENABLEOBS_COMMAND => $imsg++;
use constant DISABLEOBS_COMMAND => $imsg++;
use constant RTFC_MESSAGE => $imsg++;
use constant OBSCON_COMMAND => $imsg++;
use constant CORRSETUP_COMMAND => $imsg++;

# Error values
use constant DUPLICATE_CONNECTION => 1;

# Possible Redefine values
use constant CORR_CONECTION => 1;
use constant OBSDATA => 2;

# Possible block of data to transfer
use constant DATA_TRANSFER => 1;

sub make_command (@) {
  my $command_code = shift;

  if ($command_code==GRAB_COMMAND) {
    return new RtFC::Command::Grab(@_);
  } elsif ($command_code==ANTENNA_NAME_COMMAND) {
    return new RtFC::Command::Antenna_Name(@_);
  } elsif ($command_code==ANTENNA_ID_COMMAND) {
    return new RtFC::Command::Antenna_ID(@_);
  } elsif ($command_code==ERROR_COMMAND) {
    return new RtFC::Command::Error(@_);
  } elsif ($command_code==REDEFINE_COMMAND) {
    return new RtFC::Command::Redefine(@_);
  } elsif ($command_code==TRANSFER_COMMAND) {
    return new RtFC::Command::Transfer(@_);
  } elsif ($command_code==DATA_RECEIVED_COMMAND) {
    return new RtFC::Command::Data_Received(@_);
  } elsif ($command_code==SHOWOBS_COMMAND) {
    return new RtFC::Command::Showobs();
  } elsif ($command_code==ENABLEOBS_COMMAND) {
    return new RtFC::Command::Enableobs(@_);
  } elsif ($command_code==DISABLEOBS_COMMAND) {
    return new RtFC::Command::Disableobs(@_);
  } elsif ($command_code==RTFC_MESSAGE) {
    return new RtFC::Command::Message(@_);
  } elsif ($command_code==OBSCON_COMMAND) {
    return new RtFC::Command::Obscon(@_);
  } elsif ($command_code==CORRSETUP_COMMAND) {
    return new RtFC::Command::Corrsetup(@_);
  } else {
    croak "Unsupported command type ($command_code) at ";
  }
}

sub make_field ($$) {
  my ($subname, $field) = @_;
  my ($package, $filename, $line) = caller();

  my $sub = ("package $package;
              sub $subname {
                my \$self = shift;

		if (\@_) {
		  \$self->{$field} = shift;
		}
		return \$self->{$field};
              }");
  eval $sub;
}

sub make_commandfield ($$) {
  my ($subname, $index) = @_;
  my ($package, $filename, $line) = caller();

  my $sub = ("package $package;
              sub $subname {
                my \$self = shift;

		if (\@_) {
		  \$self->value->[$index] = shift;
		}
		return \$self->value->[$index];
              }");
  eval $sub;
}

sub make_simplecommandfield ($) {
  my ($subname) = @_;
  my ($package, $filename, $line) = caller();

  my $sub = ("package $package;
              sub $subname {
                my \$self = shift;

		if (\@_) {
		  \$self->value = shift;
		}
		return \$self->value;
              }");
  eval $sub;
}


# Message format
# 
# GRAB_COMMAND
#  Code         int     Code associated with this specific request to grab data
#  MJD          double   MJD of start of integration
#  Integration  float   Amount of data (seconds) to copy and correlate
#  Mode         string  4 character string indicating which DAS channels to
#                       copy (xxxx is all oooo is none).
#  Newcorr      byte    1 if "grab" command, 0 if "recorr" command
#  nobs         byte    Number of antenna in following array
#  Observatories array  Optional array of 2 letter observtory codes, specifying
#                       which observatories are taking part in this correlation
#
#
# DATA_RECEIVED_COMMAND
#     Command used by obs.pl child to tell parent that all the data has been
#     collected and is ready to correlate
#  code         int     Correlation identification code
#  mjd          float   MJD at start of data
#  dir          string  Path to where the baseband data is stored
#  file         string  Filename of basebande data
#
#
# OBSCON_COMMAND
#     Summary of currently connected commands
# Array of
#  Code         string  2 character antenna code
#  Name         string  Antenna name (max 20 char transmitted)
#  enabled      byte    1 if antenna is enabled fringe checking or not
#
#
# CORRSETUP_COMMAND
#     Specific parameters to use for next correlation
# nchan         int    Number of spectral points
# newcorr       byte   1 if this is a grab command. 0 if recorr
# pack          nbyte  bit encoding type
# nant          int    Number of entries in next array
# Array of
#  Ant iD       string 2 character antenna code
#  offset       float  Antenna delay offset in usec
# nant          int    Number of entries in next array
# Array of
#  Ant ID       string 2 character antenna code
#  channel      byte   Channel number to correlate


# sub command_unpack {
#   my $value = shift;
#   if (!defined $value) {
#     carp "pack: undefined value ";
#     return undef;
#   }

#   my $template;
#   my $command_code = unpack 'C', $value;

#   if ($command_code==GRAB_COMMAND) {
#     $template = 'xNdfA4CC';
#     my ($code, $mjd, $int, $mode, $newcorr, $nobs) = unpack $template, $value;
#     my @obs = ();
#     if ($nobs>0) {
#       if ($nobs*2>RtFC_COMMAND_SIZE-23) {
# 	croak "Too many connected observatories! Increase RtFC_COMMAND_SIZE!!\n";
#       }
#       $template = 'x23';
#       foreach (1..$nobs) {
# 	$template .= 'A2';
#       }
#       #warn "Unpacking GRAB using $template\n";
#       @obs = unpack $template, $value;
#     }
#     return new RtFC::Command($command_code, [$code, $mjd, $int, $mode,
# 					     $newcorr, [@obs]]);
#   } elsif ($command_code==OBSCON_COMMAND) {
#     $template = 'xC';
#     my ($nobs) = unpack $template, $value;
#     my @obssum;
#     if ($nobs>0) {
#       if ($nobs*(3+21+1)+2>RtFC_COMMAND_SIZE) {
# 	croak "Too many connected observatories! Increase RtFC_COMMAND_SIZE!!\n";
#       }
#       for (my $i=0; $i<$nobs; $i++) {
# 	$template = sprintf('x%dZ3Z21C', 2+$i*(21+3+1));
# 	push @obssum, [unpack $template, $value];
#       }
#     }
#     return new RtFC::Command($command_code, [@obssum]);

#   } elsif ($command_code==ANTENNA_NAME_COMMAND) {
#     $template = sprintf('xZ%d', RtFC_COMMAND_SIZE-1);
#     my $val = unpack $template, $value;
#     return new RtFC::Command($command_code, $val);

#   } elsif ($command_code==ANTENNA_ID_COMMAND) {
#     $template = 'xA2';
#     my $val = unpack $template, $value;
#     return new RtFC::Command($command_code, $val);

#   } elsif ($command_code==RTFC_MESSAGE) {
#     $template = 'xZ*';
#     my $msg = unpack $template, $value;
#     return new RtFC::Command($command_code, $msg);

#   } elsif ($command_code==ERROR_COMMAND || $command_code==REDEFINE_COMMAND) {
#     $template = 'xC';
#     my $val = unpack $template, $value;
#     return new RtFC::Command($command_code, $val);

#   } elsif ($command_code==TRANSFER_COMMAND) {
#     $template = 'xCx2NdA4N';
#     my ($type, $code, $mjd, $mode, $size) = unpack $template, $value;
#     return new RtFC::Command($command_code, [$type, $code, $mjd, $mode, $size]);

#   } elsif ($command_code==DATA_RECEIVED_COMMAND) {
#     $template = sprintf('xNdZ%dZ21', RtFC_COMMAND_SIZE-1-4-8-21);
#     my ($code, $mjd, $dir, $file) = unpack $template, $value;
#     return new RtFC::Command($command_code, [$code, $mjd, $dir, $file]);

#   } elsif ($command_code==SHOWOBS_COMMAND) {
#     return new RtFC::Command($command_code, undef);

#   } elsif ($command_code==ENABLEOBS_COMMAND 
# 	   || $command_code==DISABLEOBS_COMMAND) {
#     $template = sprintf('x4A2');
#     my ($obsid) = unpack $template, $value;
#     return new RtFC::Command($command_code, $obsid);

#   } elsif ($command_code==CORRSETUP_COMMAND) {
#     $template = 'xnCCC';
#     my $bytesused = 6;
#     #my ($nchan, $newcorr, $pack, $nant) = unpack $template, $value;
#     my @vals = unpack $template, $value;
#     my ($nchan, $newcorr, $pack, $nant) = @vals;
    
#     my @offsets = ();
#     my $oldbytesused = $bytesused;
#     $bytesused += $nant*(2+4);
#     if ($nant>0) {
#       if ($bytesused>RtFC_COMMAND_SIZE) {
# 	croak "Too many antenna offsets! Increase RtFC_COMMAND_SIZE!!\n";
#       }
      
#       for (my $i=0; $i<$nant; $i++) {
# 	$template = sprintf('x%da2f', $oldbytesused+$i*(2+4));
# 	push @offsets, [unpack $template, $value];
#       }
#     }

#     $template = sprintf('x%dC', $bytesused);
#     $nant = unpack $template, $value;
#     my %chans = ();
#     my $oldbytes = $bytesused;
#     $bytesused += 1 + $nant*(2+1);
#     if ($nant>0) {
#       if ($bytesused>RtFC_COMMAND_SIZE) {
# 	croak "Too many antenna channel values! Increase RtFC_COMMAND_SIZE!!\n";
#       }
      
#       for (my $i=0; $i<$nant; $i++) {
# 	$template = sprintf('x%da2C', $oldbytes+1+$i*(2+1));
# 	my ($key, $val) = unpack $template, $value;
	
# 	if ($key eq '**') {
# 	  $chans{DEFAULT} = $val;
# 	} else {
# 	  $chans{$key} = $val;
# 	}
#       }
#       if (! exists($chans{DEFAULT})) {
# 	croak 'RtFC: Internall error: Missing channel default ';
#       }
#     }

#     return new RtFC::Command($command_code, [$nchan, $newcorr, $pack, [@offsets], \%chans]);

#   } else {
#     carp "Cannot unpack this type of command ($command_code)";
#     return undef;
#   }
# }

sub recv_data ($$$$$$) {
  my ($sock, $size, $type, $sel, $killedsock, $i) = @_;

  my $buf;
  my $status = $sock->recv($buf, $size);

  if (! defined $status) {
    warn "Error reading $type socket\n";
    return undef;
  } elsif (length $buf == 0) {
    # Socket must have closed
    print "Remote $type socket closed\n";
    $sel->remove($sock) if (defined $sel);
    $sock->shutdown(2);
    $_[0] = undef; # Set $sock to undef
    if (defined $i) {
      unshift @{$killedsock}, $i; # Store largest index first
    }
    return undef;
  } else {
    return $buf;
  }
}

sub recv_command ($$$;$$) {
  my ($sock, $sel, $type, $i, $killedsock) = @_;

   print "** Receiving command from $type\n" if (DEBUG);

  my $buf = recv_data($sock, 1+4, $type, $sel, $killedsock, $i);
  return undef if (! defined $buf);

  my ($cmd_type,$size) = unpack 'CN', $buf;

  $size-= 1+4;

  my $cmd = make_command($cmd_type);
  print "**   type $cmd_type  " if (DEBUG);
  print $cmd->name, "\n" if (DEBUG);

  if ($size>0) {
    $buf = recv_data($sock, $size, $type, $sel, $killedsock, $i);
    return $buf if (!defined $buf);

    $cmd->command_unpack($buf);
  }

  return $cmd;

}

package RtFC::ObsLink;

use RtFC;

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;

  my $self = {OBS_SOCKET => shift,
	      OBS_DATASOCKET => undef,
	      OBS_IP => undef,
	      OBS_NAME => undef,
	      OBS_CODE => undef,
	      OBS_PID => undef,
	      OBS_ENABLED => 1
	     };
  bless ($self, $class);

  return $self;
}

make_field('socket', 'OBS_SOCKET');
make_field('datasocket', 'OBS_DATASOCKET');
make_field('ip', 'OBS_IP');
make_field('name', 'OBS_NAME');
make_field('code', 'OBS_CODE');
make_field('pid', 'OBS_PID');
make_field('enabled', 'OBS_ENABLED');

sub disable {
  my $self = shift;
  $self->enabled(0);
}

sub enable {
  my $self = shift;
  $self->enabled(1);
}

package RtFC::Correlation;

use Carp;
use RtFC;

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;

  my $code = shift;
  my $mjd = shift;
  my $int = shift;
  my $mode = shift;
  my $nchan = shift;
  my $pack = shift;

  my $offsets = shift;
  my $clockrates = shift;
  my $obs = shift;

  my %offsets = ();
  if (defined $offsets) {
    while (my ($key,$value) = each %{$offsets}) {
      $offsets{$key} = $value;
    }
  }

  my %clockrates = ();
  if (defined $clockrates) {
    while (my ($key,$value) = each %{$clockrates}) {
      $clockrates{$key} = $value;
    }
  }

  my %obsdata = ();
  if (defined $obs) {
    foreach (@$obs) {
      $obsdata{$_} = undef;
    }
  } 

  my $self = {CORR_CODE => $code,
	      MJD => $mjd,
	      INT => $int,
	      MODE => $mode,
	      NCHAN => $nchan,
	      PACK => $pack,
	      OFFSETS => \%offsets,
	      CLOCKRATES => \%clockrates,
	      CORR_OBS => \%obsdata,
	     };
  bless ($self, $class);

  return $self;
}

make_field('code', 'CORR_CODE');
make_field('mjd', 'MJD');
make_field('integration', 'INT');
make_field('mode', 'MODE');
make_field('nchan', 'NCHAN');
make_field('pack', 'PACK');
make_field('offsets', 'OFFSETS');
make_field('clockrates', 'CLOCKRATES');
make_field('obs', 'CORR_OBS');

package RtFC::Command;
use Carp;
use RtFC;

use constant DEBUG => 0;

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;

  my $self  = {
	       VALUE => undef
	      };
  bless ($self, $class);

  $self->setvals(@_);
  
  return $self;
}

sub setvals {
  my $self = shift;

  my $simple = $self->simple;
  if ($simple==2) {
    # Nothing to do
  } elsif ($simple) {
    $self->value(shift);
  } else {
    $self->value([@_]);
  }
}

make_field('value', 'VALUE');

sub type { return BASE_COMMAND }
sub name { return 'Base Command type'; }
sub size { return 0; }
sub simple { return 0; }
sub template { return ''; }

sub send {
  my $self = shift;
  my $sock = shift;
  my $where = shift;

  my $buf = $self->command_pack();

  printf("Sending command %s\n", $self->name) if (DEBUG);

  my $sent = $sock->send($buf);
  if (!defined $sent) {
    carp "Error sending message to $where\n";
    return 0;
  } elsif ($sent!=length($buf)) {
    carp "Only sent $sent bytes!!\n";
    return undef;
  }
  return 1;
}

# Tricky
# GRAB_COMMAND
# OBSCON_COMMAND
# ANTENNA_NAME - string with size
# RTFC_MESSAGE - string
# DATA_RECEIVED_COMMAND
# CORRSETUP_COMMAND

sub command_pack {
  my $self = shift;

  my $simple = $self->simple;
  my $template = 'CN'.$self->template;
  my $size = 1+4+$self->size;

#   if ($self->type==ANTENNA_ID_COMMAND) {
#     $template .= 'A2';
#     $size += 2;
#   } elsif ($self->type==ERROR_COMMAND) {
#     $template .= 'C';
#     $size += 1;
#     $simple = 1;
#   } elsif ($self->type==REDEFINE_COMMAND) {
#     $template .= 'C';
#     $size += 1;
#   } elsif ($self->type==TRANSFER_COMMAND) {
#     $template .= 'Cx2NdA4N';
#     $size += 1+2+4+4+4+4;
#   } elsif ($self->type==SHOWOBS_COMMAND) {
#     $simple = 2;
#   } elsif ($self->type==ENABLEOBS_COMMAND 
# 	   || $self->type==DISABLEOBS_COMMAND) {
#     $template = 'x3A2';
#     $size += 3+2;
#   } else {
#     croak "Cannot pack this type of command ";
#     return undef;
#   }

  if ($simple==2) {
    return pack $template, $self->type, $size;
  } elsif ($simple) {
    return pack $template, $self->type, $size, $self->value;
  } else {
    return pack $template, $self->type, $size, @{$self->value};
  }
}

sub command_unpack {
  my $self = shift;
  my $buf = shift;

  if (!defined $buf) {
    carp "unpack: undefined value ";
    return undef;
  }

  my $template = $self->template;
  my @vals = unpack $template, $buf;

  $self->setvals(@vals);
}

package RtFC::Command::Grab;
use Carp;
use RtFC;
use vars qw(@ISA);

use constant DEBUG => 0;


@ISA = ("RtFC::Command");

sub type { return GRAB_COMMAND }
sub name { return 'Grab' }
sub size { return 4+8+4+4+1+1 }
sub simple { return 0 }
sub template { return 'NdfA4CC' }

make_commandfield('code', 0);
make_commandfield('mjd', 1);
make_commandfield('integration', 2);
make_commandfield('mode', 3);
make_commandfield('newcorr', 4);
make_commandfield('observatories', 5);

sub command_pack {
  my $self = shift;

  my $template = 'CN'.$self->template;
  my $size = 1+4+$self->size;

  my @obs = ();
  if (defined $self->observatories) {
    @obs = @{$self->observatories};
  }
  my $nobs = scalar(@obs);
  foreach (@obs) {
    $template .= 'A2';
    $size += 2;
 }

  warn "GRAB_COMMAND: packing ($template) $size @{$self->value}[0..4] $nobs @obs\n" if (DEBUG);

  return pack $template, $self->type, $size, @{$self->value}[0..4], 
    $nobs, @obs;

}

sub command_unpack {
  my $self = shift;
  my $value = shift;

  if (!defined $value) {
    carp "unpack: undefined value ";
    return undef;
  }

  my $template = $self->template;

  my @vals = unpack $template, $value;
  my $nobs = pop @vals;

  warn "GRAB_COMMAND: Unpack ($template) @vals $nobs\n" if (DEBUG);

  my @obs = ();
  if ($nobs>0) {
    $template = sprintf 'x%d', $self->size;
    foreach (1..$nobs) {
      $template .= 'A2';
    }
    @obs = unpack $template, $value;

    warn "GRAB_COMMAND: Unpack ($template) @obs\n" if (DEBUG);
  }


  
  $self->setvals(@vals, [@obs]);

}

package RtFC::Command::Obscon;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return OBSCON_COMMAND }
sub name { return 'Obscon' }
sub size { return 1 }
sub simple { return 0 }
sub template { return 'C' }

sub command_pack {
  my $self = shift;
  
  my $template = 'CN'.$self->template;
  my $size = 1+4+$self->size;

  my $nobs = scalar(@{$self->value});
  if ($nobs>255) {
    croak sprintf("Too many (%d) antennas connected. Max is 255", $nobs);
  }

  $size += (3+21+1)*$nobs;
  my $buf = pack $template, $self->type, $size, $nobs;

  foreach (@{$self->value}) {
    $buf .= pack 'Z3Z21C', $_->[0], $_->[1], $_->[2];
  }

  return $buf;
}

sub command_unpack {
  my $self = shift;
  my $value = shift;

  if (!defined $value) {
    carp "unpack: undefined value ";
    return undef;
  }

  my $template = $self->template;
  my $nobs = unpack $template, $value;

  my @obssum;
  if ($nobs>0) {
    for (my $i=0; $i<$nobs; $i++) {
      $template = sprintf('x%dZ3Z21C', $self->size+$i*(21+3+1));
      push @obssum, [unpack $template, $value];
    }
  }
  $self->setvals(@obssum);
}

package RtFC::Command::Error;
use RtFC;
use Carp;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;

  my $self  = {
	       VALUE => shift
	      };
  bless ($self, $class);

  return $self;
}

sub type { return ERROR_COMMAND }
sub name { return 'Error' }
sub size { return 1 }
sub simple { return 1 }
sub template { return 'C' }

make_simplecommandfield('error');

package RtFC::Command::Showobs;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;

  my $self  = {};
  bless ($self, $class);

  return $self;
}

sub type { return SHOWOBS_COMMAND }
sub name { return 'Show Observatories' }
#sub size { return 0 }
sub simple { return 2 }
#sub template { return 'C' }

package RtFC::Command::Corrsetup;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return CORRSETUP_COMMAND }
sub name { return 'Corrsetup' }
sub size { return 2+1+1+1 }
sub simple { return 0 }
sub template { return 'nCCC' }

sub command_pack {
  my $self = shift;
  
  my $template = 'CN'.$self->template;
  my $size = 1+4+$self->size;

  my @offsets = ();
  if (defined $self->offsets) {
    @offsets = @{$self->offsets};
  }
  my $noffsets = scalar(@offsets);
  if ($noffsets>255) {
    croak sprintf("Too many (%d) antenna offsets. Max is 255", $noffsets);
  }
  $size += $noffsets*(2+4);

  my @clockrates = ();
  if (defined $self->clockrates) {
    @clockrates = @{$self->clockrates};
  }
  my $nrates = scalar(@clockrates);
  if ($nrates>255) {
    croak sprintf("Too many (%d) antenna rates. Max is 255", $nrates);
  }
  $size += $nrates*(2+4)+1;

  my $buf = pack($template, $self->type, $size, $self->nlags,
		 $self->newcorr, $self->pack, $noffsets);

  foreach (@offsets) {
    $buf .= pack 'a2f', $_->[0], $_->[1];
  }

  $buf .= pack 'C', $nrates;
  foreach (@clockrates) {
    $buf .= pack 'a2f', $_->[0], $_->[1];
  }

  my $xx = length($buf);

  return $buf;
}

sub command_unpack {
  my $self = shift;
  my $buf = shift;

  if (!defined $buf) {
    carp "unpack: undefined value ";
    return undef;
  }

  my $xx = length $buf;

  my $template = $self->template;

  my $size = $self->size;

  my @vals = unpack $template, $buf;
  my $nant = pop @vals;

  my @offsets = ();
  if ($nant>0) {
    for (my $i=0; $i<$nant; $i++) {
      $template = sprintf('x%da2f', $size);
      push @offsets, [unpack $template, $buf];
      $size += (2+4);
    }
  }

  $template = sprintf('x%dC', $size);
  $nant = unpack $template, $buf;
  $size += 1;

  my @clockrates = ();
  if ($nant>0) {
    for (my $i=0; $i<$nant; $i++) {
      $template = sprintf('x%da2f', $size);
      push @clockrates, [unpack $template, $buf];
      $size += (2+4);
    }
  }

  $self->setvals(@vals, \@offsets, \@clockrates);
}

make_commandfield('nlags', 0);
make_commandfield('newcorr', 1);
make_commandfield('pack', 2);
make_commandfield('offsets', 3);
make_commandfield('clockrates', 4);

package RtFC::Command::Data_Received;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return DATA_RECEIVED_COMMAND }
sub name { return 'Data Received' }
sub size { return 4+8+256+21+4 }
sub simple { return 0 }
sub template { return 'NdZ256Z21A4' }

make_commandfield('code', 0);
make_commandfield('mjd', 1);
make_commandfield('dir', 2);
make_commandfield('file', 3);
make_commandfield('mode', 4);

package RtFC::Command::Antenna_Name;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return ANTENNA_NAME_COMMAND }
sub name { return 'Antenna Name' }
sub size { return 80 }
sub simple { return 1 }
sub template { return 'Z80' }

make_simplecommandfield('antenna_name');

package RtFC::Command::Antenna_ID;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return ANTENNA_ID_COMMAND }
sub name { return 'Antenna ID' }
sub size { return 3 }
sub simple { return 1 }
sub template { return 'Z3' } # Why not A2??

make_simplecommandfield('antenna_id');

package RtFC::Command::Enableobs;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return ENABLEOBS_COMMAND }
sub name { return 'Enable Obs' }
sub size { return 2 }
sub simple { return 1 }
sub template { return 'A2' }

make_simplecommandfield('antenna_id');

package RtFC::Command::Disableobs;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return DISABLEOBS_COMMAND }
sub name { return 'Disable Obs' }
sub size { return 2 }
sub simple { return 1 }
sub template { return 'A2' }

make_simplecommandfield('antenna_id');

package RtFC::Command::Transfer;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return TRANSFER_COMMAND }
sub name { return 'Transfer' }
sub size { return 1+4+8+4+4 }
sub simple { return 0 }
sub template { return 'CNdA4N' }

make_commandfield('transfer_type', 0);
make_commandfield('code', 1);
make_commandfield('mjd',  2);
make_commandfield('mode', 3);
make_commandfield('datasize', 4);

package RtFC::Command::Redefine;
use Carp;
use RtFC;

use vars qw(@ISA);
@ISA = ("RtFC::Command");

sub type { return REDEFINE_COMMAND }
sub name { return 'Redefine' }
sub size { return 1 }
sub simple { return 1 }
sub template { return 'C' }

make_simplecommandfield('redef');

1;

