#!/usr/bin/perl -w

# **TODO**
#
# Send message back to corr saying it wont send data so things can be cleaned up
# cope (somehow) with losing connection to control.pl and corr.pl
# Fork when sending data?

use strict;

use IO::Socket;
use IO::Select;
use Getopt::Long;
use POSIX qw (floor);
use Astro::Time;

use constant RECONNECT_TIMEOUT => 5;

use RtFC;

sub process_control_command ($);
sub Shutdown();
sub open_connections ($);
sub close_connection($$);

$SIG{INT} = \&Shutdown;

system("renice 19 $$\n");

my $control_host = defined($ENV{RTFC_CONTROL})
                          ? $ENV{RTFC_CONTROL}: 'localhost';
my $corr_host = defined($ENV{RTFC_CORRELATOR})
                          ? $ENV{RTFC_CORRELATOR}: 'localhost';

my $antenna_name = defined($ENV{RTFC_ANTNAME})
                          ? $ENV{RTFC_ANTNAME}: 'Test Antenna';
my $antenna_id = defined($ENV{RTFC_ANTID}) 
                          ? $ENV{RTFC_ANTID}: 'Tt';

my $defaultpath = '.';
my $data_rate = 32;
my $swapped = 0;
my $force_mode = undef;
my $record_mode = undef;
my $prepack = undef;
my $clockoffset = 0;

GetOptions('control=s'=>\$control_host, 'correlator=s'=>\$corr_host,
	   'antname=s'=>\$antenna_name, 'antid=s'=>\$antenna_id,
	   'default=s'=>\$defaultpath, 'path=s'=>\$defaultpath,
	   'swapped'=>\$swapped, 'rate=i'=>\$data_rate, 
	   'force=s'=>\$force_mode, 'recordmode=s'=>\$record_mode,
	   'prepack=s'=>\$prepack, 'offset=i'=>\$clockoffset);

## This is where the data should be collected
#if ($defaultpath ne '.') {
#  chdir $defaultpath || die "Could not chdir to $defaultpath: $!\n";
#}

if (lc($record_mode) eq 'buggery' || lc($record_mode) eq 'jumper') {
  $record_mode = 'ooxx';
}

if (lc($record_mode) eq 'nobuggery' || lc($record_mode) eq 'nojumper') {
  $record_mode = 'oxox';
}

if (lc($record_mode) eq 's2') {
  $record_mode = 'ooxx';
}

if (defined $record_mode) {
  $record_mode =~ s/O/o/g;
  $record_mode =~ s/0/o/g;
  $record_mode =~ s/X/x/g;
}
if (defined $force_mode) {
  $force_mode =~ s/O/o/g;
  $force_mode =~ s/0/o/g;
  $force_mode =~ s/X/x/g;
}
$record_mode = undef if (defined $record_mode && $record_mode eq 'xxxx');

if (defined $prepack) {
  $prepack =~ s/0/o/g;
  $prepack =~ s/O/o/g;
  $prepack =~ s/X/x/g;
  die "Unknow prepack mode $prepack\n" if !($prepack eq 'ox' || $prepack eq 'xo' || $prepack eq 'ooxx' || $prepack eq 'xxoo');
}

my $sock_control;
my $sock_corr;
my $sock_corrdata;

my $sel = new IO::Select();

open_connections($sel);

my (@ready, $buf);

# Set the xterm titlebar
print "\033]0;RtFC $antenna_name\007";

while (1) {

  my $timeout = undef;
  if (!defined $sock_control || !defined $sock_corr 
      || !defined $sock_corrdata) {
    print "Trying to reconnect\n";
    open_connections($sel);

    if (!defined $sock_control || !defined $sock_corr 
	|| !defined $sock_corrdata) {
      $timeout = RECONNECT_TIMEOUT;
    }
  }

  while(@ready = $sel->can_read($timeout)) {
    foreach my $fh (@ready) {
      my $handled = 0;
      if($fh == $sock_control) {
	# Must have data from the control process...
	$handled = 1;
	
	my $command = recv_command($sock_control,$sel,'control');

	if (ref $command) {
	  process_control_command($command);
	} else {
	  next if (! defined $command); # There was a read error
	  if ($command==-1) { # Socket was closed remotely
	    #close_connection($sock_control, $sel);
	    warn "Remote control socket closed\n";
	  } else {
	    warn "Unknown error ($command) on correlator connection\n";
	  }
	}
      } elsif ($fh == $sock_corr) {
	# Must have data from the correlator process...
	$handled = 1;

	my $command = recv_command($sock_corr,$sel,'correlator');

	if (ref $command) {
	  printf "Cannot handle command %s from correlator\n", $command->name;
	} else {
	  next if (! defined $command); # There was a read error
	  if ($command==-1) { # Socket was closed remotely
	    #warn "Remote correlator socket closed\n";
	    #close_connection($sock_corr, $sel);
	  } else {
	    warn "Unknown error ($command) on correlator connection\n";
	  }
	}
      } elsif ($fh == $sock_corrdata) {
	# Must have data from the control process...
	$handled = 1;

	my $command = recv_command($sock_corrdata,$sel,'correlator data');

	print "sock_corrdata = $sock_corrdata\n";
	if (ref $command) {
	  printf "Cannot handle command %s from correlator data port\n", 
	    $command->name;
	} else {
	  next if (! defined $command); # There was a read error
	  if ($command==-1) { # Socket was closed remotely
	    warn "Remote correlator data socket closed\n";
	    #close_connection($sock_corrdata, $sel);
	  } else {
	    warn "Unknown error ($command) on correlator data connection\n";
	  }
	}
      } else {
	warn "Data found on unhandled socket\n";
	Shutdown();
      }

      if (!$handled) {
	warn "Did not handle message from socket $fh\n";
      }
    }
  }
  
  if (!defined $sock_control && !defined $sock_corr 
      && !defined $sock_corrdata) {
    print "Sleeping\n";
    sleep(RECONNECT_TIMEOUT);
  }
}

sub process_control_command ($) {
  my $command = shift;

  if ($command->type==GRAB_COMMAND) {
    my ($day, $month, $year, $ut) = mjd2cal($command->mjd);

    printf("Grab requested for %02d/%02d/%04d %s %.1f %s\n",
	   $day, $month, $year, turn2str($ut, 'H', 0),
	   $command->integration, $command->mode);

    grab_data($command->code, $command->mjd, $command->integration,
	      $command->mode, $command->newcorr);
  } else {
    printf STDERR "Unknown command (%d) received\n", $command->type;
  }
}

sub Shutdown () {
  $sock_control->shutdown(2) if (defined $sock_control);
  $sock_corr->shutdown(2) if (defined $sock_corr);
  $sock_corrdata->shutdown(2) if (defined $sock_corrdata);
  exit;
}

sub grab_data {
  my ($code, $mjd, $integration, $mode, $newcorr) = @_;

  $mode =~ s/0/o/g;
  $mode =~ s/O/o/g;

  my $domode = $mode;
  if (defined $force_mode) {
    $domode = $force_mode;
  }
  
  my $real_rate = $data_rate;

  # We should check the requested mode. For the moment assume the user
  # knows what they are doing. This also only works for "8 bit" mode
  # at the moment

  if (defined $record_mode && ! $force_mode) {

    # Expand $record_mode and $mode strings
    my @record_chans = split '', $record_mode;
    my @requested_chans = split '', $mode;
    
    if (scalar @record_chans !=4 || @requested_chans !=4) {
      warn "Can only use -record-mode with 8 bit MRO modes\n";
      warn "Not tranferring data\n";
      return;
    }

    my $rchan = 0;
    foreach (@record_chans) {
      $rchan++ if ($_ eq 'x');
    }
    if ($rchan eq 0) {
      warn "Not recording any data??. Not transferring then\n";
    } elsif ($rchan eq 3) {
      warn "Record mode $record_mode has unsupported 3 channels. ".
	"Not transferring data\n";
    }
    $real_rate /= 4/$rchan;  # $rchan must be 2 or 1
    print "Set rate to $real_rate\n";

    # Work out what mode we now want to run in
    $domode = '';
    my $i = 0;
    while (length($domode)<4) {
      if ($record_chans[$i] eq 'x') {
	$domode .= $requested_chans[$i];
      }

      $i++;
      $i = $i%4;
    }
    if ($domode eq 'oooo') {
      warn "No channels in common between recorded and requested. ".
	"Not transferring\n";
    }

    # What will we tell the other end
    $mode = '';
    for ($i=0; $i<4; $i++) {
      if ($record_chans[$i] eq 'x' && $requested_chans[$i] eq 'x') {
	$mode .= 'x';
      } else {
	$mode .= 'o';
      }
    }

    print "Will do compressing using $domode\n";
    print "Will say $mode\n";
  }

  my $closest_mjd = 0;
  my $closest_file = undef;

  printf "Looking for mjd $mjd\n";
  # Get a list of files in this directory
  my $status;
  opendir(DIR, $defaultpath) || die "Could not list contents of data dir $defaultpath\n";
  while (my $datafile = readdir(DIR)) {
    next if $datafile eq '..';
    next if $datafile eq '.';
    next if (! -f "$defaultpath/$datafile");
    if ($datafile =~ /^.+\.lba$/) {
      # Get the time code
      open(DATA, "$defaultpath/$datafile") 
	  || die "Could not open $datafile  for reading: $!\n";
      my $time = <DATA>;
      chomp $time;
      close(DATA) || die "Error with file $datafile: $!\n";
      if ($time =~ /^TIME (\d\d\d\d)(\d\d)(\d\d):(\d\d)(\d\d)(\d\d)$/) {
	my $year = $1;
	my $month = $2;
	my $day = $3;
	my $hour = $4;
	my $min = $5;
	my $sec = $6+$clockoffset;
	
	my $filemjd = cal2mjd($day, $month, $year,
			      (($sec/60+$min)/60 + $hour)/24);

	if ($filemjd>$closest_mjd && $filemjd<=$mjd+0.1/60/60/24) {
	  $closest_mjd = $filemjd;
	  $closest_file =  $datafile;
	}
      } else {
	warn "Unexpected time string $time\n";
      }

    } else {
      print "Ignoring $datafile\n";
    }
  }
  closedir(DIR);
  if (!defined $closest_file) {
    warn "Could not find any suitable data\n";
  } else {
    print "Closest file is $closest_file\n";
    my $offset = floor(($mjd-$closest_mjd) * 24 * 60 *60+0.5);
    printf " Offset of $offset seconds\n";

    #my ($day, $month, $year, $ut) = mjd2cal($mjd);
    #my ($sign, $hour, $minute, $second) = time2hms($ut, 'H', 0);

    my $swapopt = '';
    if ($swapped) {
      $swapopt = '-swapped ';
    }
    my $tempfile = "VSIPACK-$antenna_id.tmp";
    my $inputfile = "$defaultpath/$closest_file";

    if (defined $prepack) {
      my $thisrate = $data_rate*2;
      my $system = "vsipack -offset $offset ".
	"-int $integration -mode $prepack ".
	  "-outfile PRE-${tempfile} '$inputfile'";
      print "Running: $system\n";
      $status = system $system;
      if ($status==-1) {
	# Failed to start
	exit(0);
      } elsif (!$status) {
	warn "Not transfering data\n";
	return;
      }
      
      $inputfile = "PRE-${tempfile}";
      $offset = 0;
    }

    my $system = "vsipack -offset $offset ".
      "-int $integration -mode $domode $swapopt".
	"-outfile $tempfile '$inputfile'";

    print "Running: $system\n";
    $status = system $system;
    if ($status==-1) {
      # Failed to start
      exit(0);
    } elsif (!$status) {
      warn "Not transfering data\n";
      return;
    }
    print "Done!\n";

    if (!defined $sock_corrdata) {
      warn "Correlator data socket not open. Cannot send data\n";
      return;
    }

    # Check the size of the file
    my @stat = stat($tempfile);
    if (@stat) {
      my $size = $stat[7];

      my $command = new RtFC::Command::Transfer(DATA_TRANSFER, $code,
						$mjd, $mode, $size);

      $status = $command->send($sock_corrdata, 'correlator data');
      Shutdown() if (!defined($status));

      # Read file in chunks and send to correlator
      open(VSIDATA, $tempfile) || die "Could not open $tempfile\n";
      my $transfer_start = time;

      my $buf;
      my $bytestogo = $size;
      my $bytestoread = 1e6;
      while ($bytestogo>0) {
	$bytestoread = $bytestogo if ($bytestogo<$bytestoread);
	$status = sysread VSIDATA, $buf, $bytestoread;
	$bytestogo -= $status;
	if ($status==0) {
	  die "Error reading local file ($bytestoread)\n";
	}

	my $sent = $sock_corrdata->send($buf);
	if (!defined $sent) {
	  die "Error sending data to correlator\n";
	} elsif ($sent!=length($buf)) {
	  die "Only sent $sent bytes!!\n";
	}
      }
      my $transfer_close = time;
      
      printf("Data transfer took %.1f minutes\n",
	     ($transfer_close-$transfer_start)/60);
      close(VSIDATA);
    } else {
      warn "Failed to stat $tempfile - something must have gone wrong\n";
    }
  }
}

sub open_connections ($) {
  my $sel = shift;

  my $command_name = new RtFC::Command::Antenna_Name($antenna_name);
  my $command_id = new RtFC::Command::Antenna_ID($antenna_id);

  # Connect to the central control process

  my $status;
  if (!defined $sock_control) {
    warn "Try and open sock_control\n";
    # How messy is your drawer?
    $sock_control = IO::Socket::INET->new(PeerAddr => $control_host,
					     PeerPort => OBSPORT);

    if (defined $sock_control) {
      $status = $command_name->send($sock_control, 'control process');
      Shutdown() if (! defined $status);

      $status = $command_id->send($sock_control, 'control process');
      Shutdown() if (! defined $status);

      $sel->add($sock_control);
    }

  }

  #|| die "Couldn't create socket connection to control host ".
  #  "$control_host: $@\n";

  # Connect to the correlator process

  if (!defined $sock_corr) {
    warn "Try and open sock_corr\n";

    $sock_corr = IO::Socket::INET->new(PeerAddr => $corr_host,
				       PeerPort => CORRPORT
				      );
    if (defined $sock_corr) {
      $status = $command_name->send($sock_corr, 'correlator process');
      Shutdown() if (! defined $status);

      $status = $command_id->send($sock_corr, 'correlator process');
      Shutdown() if (! defined $status);

      $sel->add($sock_corr);
    }

    #|| die "Couldn't create socket connection to correlator host ".
    #"$corr_host: $@\n";
  }

  if (defined $sock_corr && ! defined $sock_corrdata) {
    # Connect to the correlator process again, this time for the 
    # data pipe

    warn "Try and open sock_control 2\n";

    $sock_corrdata = IO::Socket::INET->new(PeerAddr => $corr_host,
					   PeerPort => CORRPORT
					  );
    if (defined $sock_corrdata) {
      $status = $command_name->send($sock_corrdata, 'correlator process');
      Shutdown() if (! defined $status);

      $status = $command_id->send($sock_corrdata, 'correlator process');
      Shutdown() if (! defined $status);

      # Redefine this second correlator pipe as a data pipe, so the remote
      # process knows tospawn off and treat the second connection
      # differently

      my $command_redef = new RtFC::Command::Redefine(OBSDATA);
      $status = $command_redef->send($sock_corrdata, 'control process');

      #|| die "Couldn't create socket connection to $corr_host: $@\n";

      $sel->add($sock_corr);
    }
  }
}


sub close_connection($$) {
# Close a network connection
  my ($sock, $sel) = @_;

  warn "Closing socket $sock\n";
  $sel->remove($sock);

  $sock->shutdown(2);

  $_[1] = undef;

}
