#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define STRSIZE 100
#define BUFSIZE 1e6


int main (int argc, char *argv[]) {
  int  opt, status, datafile, outfile, i, expansion;
  char header[16], fname[STRSIZE+1], outname[STRSIZE+1];
  unsigned char *buf, *obuf, *iptr, *optr, tmp;
  ssize_t nread, nwrote;

  /* Options */
  int swapped;                /* IFs swapped (need to swap nybbles */
  char mode[STRSIZE+1] = "";  /* Current mode we want to extract */

  struct option options[] = {
    {"mode", 1, 0, 'm'},
    {"swapped", 0, 0, 'x'},
    {0, 0, 0, 0}
  };

  swapped = 0;
  
  while ((opt = getopt_long_only(argc, argv, "", options, NULL)) != EOF)
    switch (opt) {
    case 'm':
      if (strlen(optarg)>STRSIZE) {
	fprintf(stderr, "MODE option too long\n");
	return 0;
      }
      strcpy(mode, optarg);
      break;

    case 'x':
      swapped = 1;
      break;

    default:
      printf("Unknown option\n");
      return(0);
      break;
    }

  /* Should be just two option now */

  if (argc != optind+2) {
    fprintf(stderr, "Usage: vsipack [options] <inputfile> <outputfile>\n");
    return(0);
  }

  if (strlen(argv[optind]) > STRSIZE) {
    fprintf(stderr, "filename too long\n");
    return(0);
  } 
  if (strlen(argv[optind+1]) > STRSIZE) {
    fprintf(stderr, "output filename too long\n");
    return(0);
  } 

  strcpy(fname, argv[optind]);
  strcpy(outname, argv[optind+1]);

  /* How much is the data volume going to increase by? */
    expansion = 0;
  if (strcasecmp(mode, "xxxx")==0) {
    fprintf(stderr, "Mode xxxx: No packing performed, no need to unpack!\n");
    return(0);
  } else if (strcasecmp(mode, "ooox")==0 || strcasecmp(mode, "ooxo")==0
	     || strcasecmp(mode, "oxoo")==0 || strcasecmp(mode, "xooo")==0) {
    expansion=4;
  } else if (strcasecmp(mode, "ooxx")==0 || strcasecmp(mode, "xxoo")==0
	     || strcasecmp(mode, "oxox")==0 || strcasecmp(mode, "xoxo")==0) {
    expansion=2;
  } else {
    fprintf(stderr, "Unknown mode %s. Aborting\n", mode);
    return(0);
  }

  datafile = open(fname, O_RDONLY);
  if (datafile == -1) {
    fprintf(stderr, "Error opening %s\n", fname);
    perror(NULL);
    return(0);
  }

  status = read(datafile, header, 16);

  if (status==-1 || status==0) {
    fprintf(stderr, "Error reading %s header\n", fname);
    perror(NULL);
    return(0);
  } 

  header[15] = 0;
  printf("Reading data for time %s\n", header);

  outfile = creat(outname, S_IREAD|S_IWRITE|S_IRGRP);
  if (outfile == -1) {
    fprintf(stderr, "Error creating output file %s\n", outname);
    perror(NULL);
    return(0);
  }

  header[15] = '\n';
  status = write(outfile, header, 16);
  if (status != 16) {
    fprintf(stderr, "Error writing to output file %s\n",outname);
    perror(NULL);
    return(0);
  }

  buf = malloc(BUFSIZE);
  if (buf==NULL) {
    fprintf(stderr, "Error allocating memory\n");
    return(0);
  }

  obuf = malloc(BUFSIZE*expansion);
  if (obuf==NULL) {
    fprintf(stderr, "Error allocating memory\n");
    return(0);
  }

  while (1) {
    nread = read(datafile, buf, BUFSIZE);

    if (nread==-1) {
      fprintf(stderr, "Error reading %s header\n", fname);
      perror(NULL);
      return(0);
    } else if (nread==0) {
      break;  /* Hit EOF */
    }

    /* OK now repack the data */

    if (strcasecmp(mode, "xxxx")==0) {
      fprintf(stderr, "INTERNAL ERROR DOES NOT RUN ON MODE xxxx!!!!!\n");
      return(0);

    } else if (strcasecmp(mode, "ooox")==0 || strcasecmp(mode, "ooxo")==0 ||
	       strcasecmp(mode, "oxoo")==0 || strcasecmp(mode, "xooo")==0) {
      /* Just duplicate this data across all channels */
      
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread; i++) {
	tmp = *iptr & 0x3;
	*optr = tmp | tmp<<2 | tmp<<4 | tmp<<6;
	optr++;

	tmp = *iptr & 0xC;
	*optr = tmp>>2 | tmp | tmp<<2 | tmp<<4;
	optr++;

	tmp = *iptr & 0x30;
	*optr = tmp>>4 | tmp>>2 | tmp | tmp<<2;
	optr++;

	tmp = *iptr & 0xC0;
	*optr = tmp >> 6 | tmp>>4 | tmp>>2 | tmp;
	optr++;

	iptr++;
      }
    } else if (strcasecmp(mode, "ooxx")==0 || strcasecmp(mode, "xxoo")==0) {

      iptr = buf;
      optr = obuf;
      for (i=0; i<nread; i++) {
	tmp = *iptr & 0xF;
	*optr = tmp | tmp<<4;
	optr++;

	tmp = *iptr & 0xF0;
	*optr = tmp>>4 | tmp;
	optr++;

	iptr++;
      }
    } else if (strcasecmp(mode, "oxox")==0 || strcasecmp(mode, "xoxo")==0) {
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread; i++) {
	tmp = *iptr & 0x3;
	*optr = tmp | tmp<<2;
	tmp = *iptr & 0xC;
	*optr |= tmp<<2 | tmp<<4;
	optr++;

	tmp = *iptr & 0x30;
	*optr = tmp>>4 | tmp>>2;
	tmp = *iptr & 0xC0;
	*optr |= tmp>>2 | tmp;
	optr++;

	iptr++;
      }
    } else {
      fprintf(stderr, "Unsupported mode %s\n", mode);
      return(0);
    }

    nwrote = write(outfile, obuf, nread*expansion);
    if (nwrote != nread*expansion) {
      fprintf(stderr, "Error writing to output file\n");
      perror(NULL);
      return(0);
    }
  }
  
  free(buf);
  free(obuf);
  close(datafile);
  close(outfile);

  return(1);

}
