#ifdef __APPLE__

#define OSX

#define OPENOPTIONS O_RDONLY

#else

#define LINUX
#define _LARGEFILE_SOURCE 
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#define OPENOPTIONS O_RDONLY|O_LARGEFILE

#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <limits.h>
#include <time.h>
#include <sys/time.h>
#include <cpgplot.h>
//#include <complex.h>
#include <fftw3.h>
#include <sys/ipc.h>  /* for shared memory */
#include <sys/shm.h>
#include "sh.h"

#include "vheader.h"

#define MALLOC fftwf_malloc

#define MIN(x,y)   ((x) < (y) ? (x) : (y))

#define MAX(x,y)   ((x) > (y) ? (x) : (y))

#define MAXCHAN 8

double read_head(int file, struct tm *tm);

void doplot(int nchan, int chans[], int npoint, int bchan, int echan, 
	    int naver,  int swapped, int nx, int max, float ymax, 
	    float xvals[], double *spectrum[MAXCHAN], float plotspec[],
	    char outfile[]);

#define BLOCKSIZE 4*1024*1024

#define RATE 32
#define MAXSTR 1024

double tim(void) {
  struct timeval tv;
  double t;

  gettimeofday(&tv, NULL);
  t = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;

  return t;
}

int main (int argc, char * const argv[]) {
  int nfile, i, j, k, l, file, status, nread, opt, memsize, tmp;
  int naver, readshort, nfft, nx, ny, swapped, nplot, bytesperint, first;
  register int samp;
  char msg[MAXSTR+1], *filename, *outfile, *cptr;
  unsigned char  *mem;
  unsigned short *mem16;
  char *sample;
  short *sample16;
  off_t off, offset;
  float *in[MAXCHAN], lookup[4], *plotspec, *xvals, ftmp;
  fftwf_complex *out[MAXCHAN];
  double *spectrum[MAXCHAN], t0, t1, tt;
  long long lcount[4], tcount;
  fftwf_plan plans[MAXCHAN]; /* Do up to MAXCHAN subbands at the same time */

  int verbose;       /* Prints whats happening */
  int online;        /* Run continously on live data */
  int pause;         /* Wait between subplots */
  int dump;          /* Write output spectrum */
  int domax;         /* Just print maximum value (used for blanking tests) */
  int bits;          /* Number of bits/sample */
  int nchan;         /* Total number of channels */
  int bandwidth;     /* Observing bandwidth (different byte format for 32 and 
			64 MHz data*/
  int nskip;         /* Number of bytes to skip at start of file */
  int npoint;        /* Number of spectral channels */
  int nint;          /* Integration time, in samples */
  int chans[MAXCHAN];/* Channels to correlate */
  int bchan;         /* First spectral point to plot */
  int echan;         /* Last spectral point to plot */
  int docommand;     /* Command string passed? */
  float ymax;        /* Maximum yvalue of plot */
  encodingtype encoding; /* VLBA or AT bit encoding (offset binary or sign/mag) */
  char pgdev[MAXSTR+1] = "/xs"; /* Pgplot device to use */
  char command[MAXSTR+1] = ""; /* Pgplot device to use */

  int chan1, chan2;
  vhead *header;

  /* Shared memory unique identifier (key) and the returned reference id. */
  key_t shKey;
  int shId = -1;
  ptSh sh;

  struct option options[] = {
    {"vlba", 0, 0, 'v'},
    {"npoint", 1, 0, 'n'},
    {"chan", 1, 0, 'C'},
    {"nint", 1, 0, 'N'},
    {"skip", 1, 0, 's'},
    {"if1", 1, 0, '1'},
    {"if2", 1, 0, '2'},
    {"sp1", 1, 0, 'q'},
    {"sp2", 1, 0, 'r'},
    {"ymax", 1, 0, 'y'},
    {"pause", 0, 0, 'p'},
    {"dump", 0, 0, 'D'},
    {"max", 0, 0, 'm'},
    {"command", 1, 0, 'x'},
    {"device", 1, 0, 'd'},
    {"pgdev", 1, 0, 'd'},
    {"verbose", 0, 0, 'V'},
    {"online", 0, 0, 'o'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  verbose = 0;
  online = 0;
  pause = 0;
  dump = 0;
  domax = 0;
  nskip = 0;
  docommand = 0;
  npoint = 128;
  encoding = AT;
  chan1 = -1;
  chan2 = -1;
  nint = -1;
  for (k=0;k<MAXCHAN;k++) chans[k]=0;
  bchan = -1;
  echan = -1;
  ymax = 0.0;
  sh = NULL;
  filename = NULL;

  /* Avoid compiler complaints */
  memsize = 0;
  nx = 0;
  swapped = 0;
  bytesperint = 0;
  mem = 0;
  bits = 0;
  nchan = 0;
  bandwidth = 0.0;

  nfft = 0;
  /* Read command line options */
  while (1) {
    opt = getopt_long_only(argc, argv, "d:N:s:y:n:c:C:Vpvo1:2:", 
			   options, NULL);
    if (opt==EOF) break;

    switch (opt) {
      
    case 'C':
    case 'c':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -chan option %s\n", optarg);
      else {
	if (tmp>MAXCHAN) {
	  fprintf(stderr, "Chan %s too large\n", optarg);
	} else if (tmp<=0) {
	  fprintf(stderr, "Chan %s too small\n", optarg);
	} else {
	  chans[tmp-1] = 1;
	}
      }
      break;

    case '1':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -1chan option %s\n", optarg);
      else 
	chan1 = tmp-1;
      break;

    case '2':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -2chan option %s\n", optarg);
      else 
	chan2 = tmp;
      break;

    case 'q':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -sp1 option %s\n", optarg);
      else 
	bchan = tmp;
      break;

    case 'r':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -sp2 option %s\n", optarg);
      else 
	echan = tmp;
      break;

    case 'y':
      printf("Got ymax = %s\n", optarg);
      status = sscanf(optarg, "%f", &ftmp);
      if (status!=1)
 	fprintf(stderr, "Bad -ymax option %s\n", optarg);
      else 
	ymax = ftmp;
      break;

    case 's':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -skip option %s\n", optarg);
      else 
	nskip = tmp;
      break;

    case 'n':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -npoint option %s\n", optarg);
      else 
	npoint = tmp;
      break;

    case 'N':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -nint option %s\n", optarg);
      else 
	nint = tmp;
      break;

    case 'd':
      if (strlen(optarg)>MAXSTR) {
	fprintf(stderr, "PGDEV option %s too long\n", optarg);
	return 0;
      }
      strcpy(pgdev, optarg);
      break;

    case 'x':
      if (strlen(optarg)>MAXSTR) {
	fprintf(stderr, " -command option %s too long\n", optarg);
	return 0;
      }
      strcpy(command, optarg);
      docommand = 1;
      break;

    case 'V':
      verbose = 1;
      break;

    case 'o':
      online = 1;
      break;

    case 'p':
      pause = 1;
      break;

    case 'D':
      dump = 1;
      break;

    case 'm':
      domax = 1;
      break;

    case 'v':
      encoding = VLBA;
      break;

    case 'h':
      printf("Usage: fauto [options]\n");
      printf("  -C/-chan <n>        Channel to correlate (can specify "
	     "multiple times)\n");
      printf("  -n/-npoint <n>      # spectral channels\n");
      printf("  -v/-vlba            Calculate statistics for VLBA data\n");
      printf("  -N/-init <val>      Number of ffts to average per "
	     "integration\n");
      printf("  -device <pgdev>     Pgplot device to plot to\n");
      printf("  -if1/-1 <n>         1st channel to correlate in range\n");
      printf("  -if2/-2 <n>         Last channel to correlate in range\n");
      printf("  -sp1 <n>            First spectral point to plot\n");
      printf("  -sp2 <n>            Last spectral point to plot\n");
      printf("  -s/-skip <n>        Skip this many bytes at start of file\n");
      printf("  -p/-pause           Pause between fft subintegrations\n");
      printf("  -D/-dump            Write ascii copy of spectrum\n");
      printf("  -V/-verbose         Verbose\n");
      printf("  -h/-help            This list\n");
      return(1);
      break;

    case '?':
    default:
      break;
    }
  }

  if (docommand && ! dump) {
    fprintf(stderr, "Can only run command when writing spectra!\n");
    docommand = 0;
  }


  /* Initiate plotting arrays */

  xvals = MALLOC(sizeof(float)*npoint);
  for (j=0; j<npoint; j++) {
    xvals[j] = j+0.5;
  }
  plotspec = MALLOC(sizeof(float)*npoint);

  /* Setup the lookup table */
  if (encoding==VLBA) {
    lookup[0] = +3.0;
    lookup[1] = +1.0;
    lookup[2] = -1.0;
    lookup[3] = -3.0;
  } else { /* AT */
    lookup[0] = +1.0;
    lookup[1] = -1.0;
    lookup[2] = +3.0;
    lookup[3] = -3.0;
  }

  if (online) {
  /* Create and initialize shared memory. */
    shKey = fourCharLong('v','s','i','b');

    shId = shmget(shKey, 0, 0);
    if (shId==-1) {
      perror("shmget failed"); 
      exit(1);
    }
      
    sh = (ptSh)shmat(shId, NULL, 0);
    if (sh == (void *)-1 ) {
      perror("shmat failed");
      exit(1);
    }

    filename = (char *)malloc(SH_MAXSTR); 
    if (filename==NULL) {
      perror("Allocating memory");
      exit(1);
    }
    strcpy(filename,"");
  }

  if (dump) {
    outfile = (char *)malloc(MAXSTR+1);
    if (outfile==NULL) {
      perror("Allocating memory");
      exit(1);
    }
  } else {
    outfile = NULL;
  }

  tt = 0; /* Total time */
  naver = 0;
  first = 1;

  nfile = optind;
  header = newheader();

  while (1) { // Loop for online mode
    while (1) { // Loop over (and accumulate) input files

      if (online) {
	if (strcmp(filename,"")==0 || strcmp(sh->lastfile,filename)!=0) {
	  strcpy(filename, sh->lastfile);
	} else {
	  sleep(1);
	  continue;
	}
	
      } else {
	if (nfile >= argc) break;
	filename = argv[nfile];
	nfile++;
      }

      file = open(filename, OPENOPTIONS);
      if (file==-1) {
	sprintf(msg, "Failed to open input file (%s)", filename);
	perror(msg);
	//fftwf_free(mem);
	return(1);
      }
      printf("%s\n", filename);

      /* Read the header */
      resetheader(header);
      readheader(header, file);

      if (first) {
	// Assume all files the same, for now
	printf("Reading header\n");
	nchan = header->nchan;
	bits = header->numbits;
	bandwidth = header->bandwidth;
	
	/* Check too high channel hasn't been selected */
	for (k=nchan; k<MAXCHAN; k++) {
	  if (chans[k]) {
	    fprintf(stderr, "Channel %d is too large for this mode. Ignoring\n", k+1);
	    chans[k] = 0;
	  }
	}
	
	if (chan1>0 || chan2>0) {
	  if (chan1<=0) 
	    chan1=1;
	  else if (chan1>nchan) {
	    fprintf(stderr, "Channel %d too large for current mode. Setting to %d\n",
		    chan1, nchan);
	    chan1=nchan;
	  }
	  
	  if (chan2<=0) 
	    chan2=nchan;
	  else if (chan2>nchan) {
	    fprintf(stderr, "Channel %d too large for current mode. Setting to %d\n",
		    chan2, nchan);
	    chan2=nchan;
	  }
	  if (chan2<chan1) {
	    tmp = chan1;
	    chan1 = chan2;
	    chan2 = tmp;
	  }
	  for (k=chan1-1;k<chan2;k++) {
	    chans[k] = 1;
	  }
	}
	
	/* Have any channels been selected? */
	nplot = 0;
	for (k=0; k<nchan; k++) {
	  if (chans[k]) {
	    nplot++;
	  }
	}
	if (nplot==0) {
	  fprintf(stderr, "Plotting all channels\n");
	  
	  for (k=0; k<nchan; k++) {
	    chans[k] = 1;
	  }
	  nplot=nchan;
	}

	if (bchan<0) bchan = 0;
	if (echan<0) echan = npoint-1;
	if (bchan > npoint-1) bchan = npoint-1;
	if (echan > npoint-1) echan = npoint-1;
	if (bchan==echan) {
	  fprintf(stderr, "Warning: Odd channel range. Resetting\n");
	  bchan = 0;
	  echan = npoint-1;
	}
  
	if (bchan>echan) {
	  swapped = bchan;
	  bchan = echan;
	  echan = swapped;
	  swapped = 1;
	  fprintf(stderr, "Swapping frequency axis\n");
	} 
	echan++;
	
	/* Bytes per integration */
	if (bits==10) {
	  bytesperint = npoint*2*16*nchan/8;
	} else {
	  bytesperint = npoint*2*bits*nchan/8;
	}

	printf("Bytes per integration = %d\n", bytesperint);
	
	memsize = floor(BLOCKSIZE/bytesperint)*bytesperint+0.1;

	mem = MALLOC(memsize);
	if (mem==NULL) {
	  sprintf(msg, "Trying to allocate %d byes", memsize);
	  perror(msg);
	  return(1);
	}

	/* Initiate fftw */
	
	for (k=0; k<nchan; k++) {
	  if (!chans[k]) continue;
	  in[k] = MALLOC(sizeof(float) * npoint*2);
	  if (in[k]==NULL) {
	    perror("Allocating memory");
	    exit(1);
	  }
	  out[k] = MALLOC(sizeof(fftwf_complex) *npoint);
	  if (out[k]==NULL) {
	    perror("Allocating memory");
	    exit(1);
	  }
	  spectrum[k] = MALLOC(sizeof(double)*npoint);
	  if (spectrum[k]==NULL) {
	    perror("Allocating memory");
	    exit(1);
	  }
	  for (j=0; j<npoint; j++) {
	    spectrum[k][j] = 0;
	  }

	  plans[k] = fftwf_plan_dft_r2c_1d(npoint*2, in[k], out[k], 0);
	}

	/* How many plots to do */;
	ny = ceil(sqrt(nplot));
	nx = (nplot+ny-1)/ny;
	
	if (!domax) {
	  if (cpgbeg(0,pgdev,nx,ny)!=1) {
	    fprintf(stderr, "Error calling PGBEGIN");
	    return(1);
	  }
	  cpgask(0);
	}
      }
    
      for (i=0; i<4; i++) {
	lcount[i] = 0;
      }

      offset = 0;
      readshort = 0;

      if (nskip>0) {
	offset = lseek(file, nskip, SEEK_CUR);
	if (offset==-1) {
	  perror(msg);
	  //fftwf_free(mem);
	  return(1);
	}
      }

      while (1) {
	nread = read(file, mem, memsize);
	
	if (nread==0) {
	  if (offset<=nskip) {
	    fprintf(stderr, "No data read from %s\n", filename);
	    //fftwf_free(mem);
	    close(file);
	    return(1);
	  }
	  close(file);
	  break;
	} else if (nread==-1) {
	  perror("Error reading file");
	  close(file);
	  //fftwf_free(mem);
	  return(1);
	} else if (nread%bytesperint != 0) { /* Need to read multiple of lags */
	  /* Data may be pack with multiple samples/byte or mutiple 
	     bytes/sample */
	  int shift;
	  shift = nread % bytesperint;
	  
	  off = lseek(file, -shift, SEEK_CUR); 
	  if (off==-1) {
	    perror(msg);
	    //fftwf_free(mem);
	    return(1);
	  }
	  nread -= shift;
	}
	offset += nread;
	t0 = tim();
      
	if (nread<bytesperint) { /* Not read enough */
	  if (readshort) {
	    /* Probably at end of file */
	    if (offset < bytesperint) {
	      fprintf(stderr, "Warning: Short file\n");
	      exit(1);
	    } else {
	      break; 
	    }
	  } else {
	    readshort = 1; /* Try one more time, just in case */
	    continue;
	  }
	}

	mem16 = (unsigned short*)mem;

	/* The following loops decode the data into a set of float arrays */

	
	/* Copy data into "in" array, fft then accumulate */
	for (i=0; i<nread/bytesperint; i++) {

	  if (bits==2) {
	    if (bandwidth==64) { /* 64 MHz needs to be handle separately */
	      if (nchan==2) {
		for (j=0; j<npoint/2; j++) {
		  for (k=0; k<nchan; k++) {
		    if (!chans[k]) continue;
		    samp = mem[i*bytesperint+j*2+k];
		    for (l=0; l<4; l++) {
		      in[k][j*4+l] = lookup[(samp>>(l*2))&0x3];
		      lcount[(samp>>(l*2))&0x3]++;
		    }
		  }
		}
	      } else if (nchan==1) {
		for (j=0; j<npoint/2; j++) {
		  samp = mem[i*bytesperint+j];
		  for (l=0; l<4; l++) {
		    in[0][j*4+l] = lookup[(samp>>(l*2))&0x3];
		    lcount[(samp>>(l*2))&0x3]++;
		  }
		}
	      } else {
		fprintf(stderr, "64 MHz %d channels not supported\n", nchan);
		exit(1);
	      }
	    
	    } else if (bandwidth==32) {
	      if (nchan==2) {
		for (j=0; j<npoint; j++) {
		  samp = mem[i*bytesperint+j];
		  for (k=0; k<nchan; k++) {
		    if (!chans[k]) continue;
		    in[k][j*2] = lookup[(samp>>(k*2))&0x3];
		    in[k][j*2+1] = lookup[(samp>>(k*2+2))&0x3];
		    lcount[(samp>>(k*2))&0x3]++;
		    lcount[(samp>>(k*2+2))&0x3]++;
		  }
		}
	      } else {
		fprintf(stderr, "32 MHz %d channels not supported\n", nchan);
		exit(1);
	      }

	    } else { // Normal modes. Bandwidth unimportant
	      if (nchan==4) {
		for (j=0; j<npoint*2; j++) {
		  samp = mem[i*npoint*2+j];
		  for (k=0; k<nchan; k++) {
		    if (!chans[k]) continue;
		    in[k][j] = lookup[(samp>>(k*2))&0x3];
		    lcount[(samp>>(k*2))&0x3]++;
		  }
		}
		
	      } else if (nchan==8) {
		for (j=0; j<npoint*2; j++) {
		  samp = mem16[i*npoint*2+j];
		  for (k=0; k<nchan; k++) {
		    if (!chans[k]) continue;
		    in[k][j] = lookup[(samp>>(k*2))&0x3];
		    lcount[(samp>>(k*2))&0x3]++;
		  }
		}
	      } else if (nchan==2) {
		for (j=0; j<npoint; j++) {
		  samp = mem[i*bytesperint+j];
		  for (k=0; k<nchan; k++) {
		    if (!chans[k]) continue;
		    in[k][j*2] = lookup[(samp>>(k*2))&0x3];
		    in[k][j*2+1] = lookup[(samp>>(k*2+4))&0x3];
		    lcount[(samp>>(k*2))&0x3]++;
		    lcount[(samp>>(k*2+4))&0x3]++;
		  }
		}
	      } else if (nchan==1) {
		for (j=0; j<npoint/2; j++) {
		  samp = mem[i*bytesperint+j];
		  
		  in[0][j*4] = lookup[samp&0x3];
		  in[0][j*4+1] = lookup[(samp>>2)&0x3];
		  in[0][j*4+2] = lookup[(samp>>4)&0x3];
		  in[0][j*4+3] = lookup[(samp>>6)&0x3];
		  
		  lcount[samp&0x3]++;
		  lcount[(samp>>2)&0x3]++;
		  lcount[(samp>>4)&0x3]++;
		  lcount[(samp>>6)&0x3]++;
		}
	      } else {
	      fprintf(stderr, "Unsupported number of channels %d\n", nchan);
	      exit(1);
	      }
	    }
	    
	  } else if (bits==8) {
	    sample =  (char*)&mem[i*bytesperint];
	    sample--;
	    for (j=0; j<npoint*2; j++) {
	      for (k=0; k<nchan; k++) {
		sample++;
		if (!chans[k]) continue;
		in[k][j] = *sample;
	      }
	    }
	  } else if (bits==10) {
	    sample16 =  (short*)&mem16[i*bytesperint/2];
	    sample16--;
	    for (j=0; j<npoint*2; j++) {
	      for (k=0; k<nchan; k++) {
		sample16++;
		if (*sample16&0x200) // 10th bit set
		  *sample16 |= 0xFC00; // Force upper 6 bits high;
	      else
		*sample16 &= 0x03FF; // Force upper 6 bits low;
		
		if (!chans[k]) continue;
		in[k][j] = *sample16;
	      }
	    }
	  } else {
	    fprintf(stderr, "Unsupported number of bits %d\n", bits);
	    exit(1);
	  }
	  
	  for (k=0; k<nchan; k++) {
	    if (!chans[k]) continue;
	    
	    fftwf_execute(plans[k]);
	    nfft++;
	    for (j=0; j<npoint; j++) {
	      spectrum[k][j] += out[k][j][0]*out[k][j][0] 
		+ out[k][j][1]*out[k][j][1];
	    }
	  }
	  naver++;
	  
	  if (nint>0 && naver>=nint) {
	    char tmpstr[MAXSTR];
	    
	    if (!first && pause) { /* Pause between plots */
	      printf("<Type Return to continue>\n");
	      fgets(tmpstr, MAXSTR, stdin);
	    }
	    doplot(nchan, chans, npoint, bchan, echan, naver, swapped, 
		   nx, domax, ymax, xvals, spectrum, plotspec, NULL);
	    
	    naver = 0;
	    first = 0;
	    for (k=0; k<nchan; k++) {
	      if (!chans[k]) continue;
	      for (j=0; j<npoint; j++) {
		spectrum[k][j] = 0;
	      }
	    }
	  }
	}
	t1 = tim();
	tt += t1-t0;
      } // Loop over file

      if (online) break;
    } // Loop between files


    printf("Total computational time= %0.1f seconds for %d ffts\n", 
	   tt, nfft);

    if (nint<0 || (float)naver/(float)nint>0.25) {

      if (dump) {
	// Remove leading path if present
	cptr = rindex(filename, '/');
	if (cptr==NULL) cptr = filename;
	strcpy(outfile, filename);

	// Remove trailing prefix, if present
	cptr = rindex(outfile, '.');
	if (cptr!=NULL) *cptr = 0;
	strcat(outfile, ".spec");

	printf("***** %s\n", outfile);
      } 

      doplot(nchan, chans, npoint, bchan, echan, naver, swapped, nx,	   
	     domax, ymax, xvals, spectrum, plotspec, outfile);
      if (online) {
	tt = 0;
	nfft = 0;
	naver = 0;
	first = 0;
	for (k=0; k<nchan; k++) {
	  if (!chans[k]) continue;
	  for (j=0; j<npoint; j++) {
	    spectrum[k][j] = 0;
	  }
	}
      }
      if (docommand) {
	char cmdstr[MAXSTR+2];
	if (strlen(command)+strlen(outfile)>MAXSTR) {
	  fprintf(stderr, "Outfile and commnd strig too long\n");
	} else {
	  sprintf(cmdstr, command, outfile);
	  printf("Running: %s\n", cmdstr);
	  system(cmdstr);
	}
      }
    }

    if (!online) break;

  } // Loop between files for online mode

  fftwf_free(mem);
  for (k=0; k<nchan; k++) {
    if (!chans[k]) continue;
    //fftwf_destroy_plan(plans[k]);
    //fftwf_free(in[k]); 
    //fftwf_free(out[k]);
  }
  fftwf_free(xvals);

  if (bits==2) {
    tcount = 0;
    for (i=0; i<4; i++) {
      tcount += lcount[i];
    }
    printf("Total counts %.1fe6\n", tcount/1e6);
    for (i=0; i<4; i++) {
      printf(" %5.1f", lcount[i]/(double)tcount*100);
    }
    printf("\n");
  }

  if (!domax) cpgend();

  return(1);
}

void doplot (int nchan, int chans[], int npoint, int bchan, int echan, 
	     int naver, int swapped, int nx, int domax, float ymax, 
	     float xvals[], double *spectrum[MAXCHAN], float plotspec[],
	     char outfile[]) {
  int j, k, m, q;
  float max, min, delta;
  char msg[MAXSTR+1];
  FILE *os = 0;

  if (outfile!=NULL) {
    os=fopen(outfile,"w");
    if (os==NULL) {
      fprintf(stderr, "Error opening output file %s\n", outfile);
      outfile = NULL;
    }
  }

  m = 0;
  for (k=0; k<nchan; k++) {
    if (!chans[k]) continue;
    
    max = spectrum[k][bchan] / naver/npoint/2/M_PI;
    min = max;
    
    //sum = 0;
    //for (j=0;j<npoint;j++) {
    //  sum += spectrum[k][j] / naver;
    //}
    //printf("Chan %d: Av = %f\n", k, sum/npoint/npoint/2/M_PI);

    for (j=bchan; j<echan; j++) {
      plotspec[j] = spectrum[k][j] / naver/npoint/2/M_PI;
      if (plotspec[j]>max) 
	max = plotspec[j];
      else if (plotspec[j]<min) 
	min = plotspec[j];
    }

    if (domax) {
      printf("%.1f\n", max);
      return;
    }

    delta = (max-min)*0.05;
    min -= delta/2;
    max += delta;
    if (ymax != 0.0) max = ymax;

    cpgsci(7);

    cpgbbuf();
    cpgpanl(m%nx+1,m/nx+1);
    cpgeras();
    cpgvstd();

    if (swapped) 
      cpgswin(echan,bchan,min,max);
      else 
      cpgswin(bchan,echan,min,max);
  
    cpgbox("BCNST", 0.0, 0, "BCNST", 0.0, 0);

    sprintf(msg, "Channel %d", k+1);
    if (outfile!=NULL)
      fprintf(os,"%d  (number of channels for IF Channel %d)\n", 
	      echan-bchan,k+1);
    cpglab("Channel", "Correlation", msg);
    cpgsci(2);
    cpgline(echan-bchan, &xvals[bchan], &plotspec[bchan]);

    if (outfile!=NULL)
      for (q=bchan;q<echan;++q)
	fprintf(os,"%5.0f    %f\n",xvals[q]-0.5,plotspec[q]);

    cpgebuf();

    m++;
  }
  if (outfile!=NULL)
    fclose(os);

}
