/* sff.c -- Mk4/VLBA software formatter, formatting subroutine templates. */

/*
 * $Log: sff.c,v $
 * Revision 1.4  2006/06/22 17:06:51  amn
 * Mods to compile with gcc-4.0 and probably 64-bit SuSE.
 *
 * Revision 1.3  2005/01/19 07:19:02  amn
 * Dec-2004 JIVE software formatter changes; vsib.c seek support.
 *
 * Revision 1.1  2004/08/02 18:35:39  amn
 * First release of sf.c, software Mk4/VLBA formatter.
 *
 *
 */

/* Copyright (C) 2001--2004 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


/* The 64/32/16/8 bit data type is selected by a predefined macro */
/* 'NM(a)' which '##'-catenates a suitable suffix like '64' or '8'. */
/* 'BITS' is another macro used in for loops. */

/* CRC functions update CRC register (stored at 'cr') */
/* with 'datawords' count of data words starting at 'd'. */
void
NM(crc16)(NM(ptD) cr, NM(ptD) d, int datawords) 
{
  int i;

  /* 'cr' CRC register is assumed to be initialized */
  /* with 16 data words. */
  /* (Init to all zero; or use old regs to continue. */
  for (i=0; i<datawords; i++) {
    NM(tD) n0  = *d++ ^ cr[15];  /* x^16 + */
    NM(tD) n15 = n0 ^ cr[14];  /* x^15 + */
    NM(tD) n2  = n0 ^ cr[1];   /* x^2 + 1 */

    cr[15] = n15;
    cr[14] = cr[13];
    cr[13] = cr[12];
    cr[12] = cr[11];
    cr[11] = cr[10];
    cr[10] = cr[ 9];
    cr[ 9] = cr[ 8];
    cr[ 8] = cr[ 7];
    cr[ 7] = cr[ 6];
    cr[ 6] = cr[ 5];
    cr[ 5] = cr[ 4];
    cr[ 4] = cr[ 3];
    cr[ 3] = cr[ 2];
    cr[ 2] = n2;
    cr[ 1] = cr[ 0];
    cr[ 0] = n0;
  }  /* for every data word (with multiple bits) in input data stream */
  /* The result is in 'cr[]', MSB in last data word. */
}  /* crc16 */

void
NM(crc12)(NM(ptD) cr, NM(ptD) d, int datawords) 
{
  int i;

  /* 'cr' CRC register is assumed to be initialized */
  /* with 12 data words. */
  /* (Init to all zero; or use old regs to continue. */
  for (i=0; i<datawords; i++) {
    NM(tD) n0  = *d++ ^ cr[11];  /* x^12 + */
    NM(tD) n11 = n0 ^ cr[10];  /* x^11 + */
    NM(tD) n3  = n0 ^ cr[2];   /* x^3 + */
    NM(tD) n2  = n0 ^ cr[1];   /* x^2 + */
    NM(tD) n1  = n0 ^ cr[0];   /* x + 1 */

    cr[11] = n11;
    cr[10] = cr[ 9];
    cr[ 9] = cr[ 8];
    cr[ 8] = cr[ 7];
    cr[ 7] = cr[ 6];
    cr[ 6] = cr[ 5];
    cr[ 5] = cr[ 4];
    cr[ 4] = cr[ 3];
    cr[ 3] = n3;
    cr[ 2] = n2;
    cr[ 1] = n1;
    cr[ 0] = n0;
  }  /* for every data word (with multiple bits) in input data stream */
  /* The result is in 'cr[]', MSB in last data word. */
}  /* crc12 */

/* CRC function pointer type def. */
typedef void (*NM(pCRCfunc))(NM(ptD) cr, NM(ptD) d, int datawords);



/* Transpose template bytes bit at a time (starting from MSB) */
/* into 'target' data words, either one target bit stream at a time */
/* or all bits together.  Returns ptr of next unprocessed vacant data word. */
NM(ptD)
NM(fmt)(NM(ptD) target, ptByte from, int bytes,
    int whichbit, int whichparity, NM(ptD) crcBlock, NM(pCRCfunc) pcrcf
)
{
  NM(tD) mask;
  int i, j;

  if (whichbit < 0) {
    /* Change all parallel bit streams of data. */
    mask = ~(0);  /* 0xffffffff... */
  } else {
    /* Change only a single bit stream. */
    assert( whichbit < sizeof(NM(tD))*8 );
    mask = ((NM(tD))1) << whichbit;
  }
  for (i=0; i<bytes; i++) {
    NM(tD) parity;
    tByte f;

    /* <0 even (for sync word), ==0 none, >0 odd (for all other data). */
    if (whichparity > 0) {
      parity = mask;
    } else {
      parity = 0;
    }

    f = from[i];
    /* Put every bit of source template into a new bit stream target word. */
    for (j=0; j<=7; j++) {
      NM(tD) d;

      d = ( (f & 0x80) ? mask : 0 );
      parity ^= d;
      if (whichbit < 1) {  /* all==-1 or first==0, replace/zero all */
	*target = d;
      } else {
	/* Single-bits 1--63, just add a single bit. */
	*target |= d;
      }
      if (crcBlock) {  /* used only with -1==all or 31/63==last bit */
	(*pcrcf)(crcBlock, target, 1);
      }
      target++;
      f <<= 1;  /* next bit */
    }  /* for every bit in template from byte */

    /* For both even and odd parity, add parity bit data word. */
    if (whichparity != 0) {
      if (whichbit < 1) {  /* all==-1 or first==0, replace/zero all */
	*target++ = (parity & mask);
      } else {
	/* Single-bits 1--63, just add a single bit. */
	*target++ |= (parity & mask);
      }
    }
  }  /* for every target bit */
#if 0
  printf("fmt: %08X after inserting %d bytes\n", (unsigned int)target, bytes);
#endif
  return (target);
}  /* fmt */


NM(ptD)
NM(mk4fmtc)(NM(ptD) target, NM(ptD) crcBlock, int useParity)
{
  int trk;
  unsigned char fts[14];  /* formatter time as string */

  /* Init CRC generator to all zeroes. */
  memset(crcBlock, 0, 12*sizeof(NM(tD)));

  /* Dummy headstack positions. */
  mf.auxhs1[0] = 0x55; mf.auxhs1[1] = 0x55;
  mf.auxhs2[0] = 0x55; mf.auxhs2[1] = 0x55;
  target = NM(fmt)(target, (ptByte)(&(mf.auxhs1)), 4, -1,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc12)));

  /* Run track information separately for all bits in data word. */
  for (trk=0; trk<=(BITS-2); trk++) {
    mf.auxtrk = trk;
    mf.auxsrc = trk;
    (void)NM(fmt)(target, (ptByte)(&(mf.auxtrk)), 2, trk,
	      useParity?/*odd*/1:0, NULL, NULL);
  }  /* for each track */
  /* Calculate CRC only when the last bit/track 31/63 is in place. */
  mf.auxtrk = trk;
  mf.auxsrc = trk;
  target = NM(fmt)(target, (ptByte)(&(mf.auxtrk)), 2, trk,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc12)));
  mf.auxsts = 0x00;  /* no formatter errors */
  mf.auxsid = 0x12;  /* system ID */
  target = NM(fmt)(target, (ptByte)(&(mf.auxsts)), 2, -1,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc12)));

  mf.sync[0] = 0xff; mf.sync[1] = 0xff; mf.sync[2] = 0xff; mf.sync[3] = 0xff;
  target = NM(fmt)(target, (ptByte)(&(mf.sync)), 4, -1,
	       useParity?/*even*/-1:0, crcBlock, &(NM(crc12)));

  /* Generate formatter time in BCD up to seconds. */
  assert( sprintf((char *)fts, "%04d%03d%02d%02d%02d",
		  ft.yyyy, ft.doy, ft.hh, ft.min, ft.ss
		  ) <= ((int)sizeof(fts)-1) );
  mf.timec[0] = ((fts[ 3] & 0x0f) << 4) | (fts[ 4] & 0x0f);  /* YD */
  mf.timec[1] = ((fts[ 5] & 0x0f) << 4) | (fts[ 6] & 0x0f);  /* DD */
  mf.timec[2] = ((fts[ 7] & 0x0f) << 4) | (fts[ 8] & 0x0f);  /* HH */
  mf.timec[3] = ((fts[ 9] & 0x0f) << 4) | (fts[10] & 0x0f);  /* MM */
  mf.timec[4] = ((fts[11] & 0x0f) << 4) | (fts[12] & 0x0f);  /* SS */
  target = NM(fmt)(target, (ptByte)(&(mf.timec)), 5, -1,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc12)));
  /* 'crcBlock' contains "seed" of CRC state this far. */
  return (target);
}  /* mk4fmtc */

NM(ptD)
NM(mk4fmtf)(NM(ptD) target, NM(ptD) crcBlock, int useParity)
{
  NM(tD) parity;
  NM(tD) mask = ~(0);  /* 0xffffffff... */
  tByte f100;
  tByte f8;
  tByte f;
  int j;

  /* First byte, odd parity, sub-second time code. */
  parity = mask;
  /* 0.1 and 0.01 seconds, in BCD. */
  f100 = ft.f800 / 8;
  f = (((f100 / 10) & 0x0f) << 4) | ((f100 % 10) & 0x0f) ;
  for (j=0; j<=7; j++) {
    NM(tD) d;

    d = ( (f & 0x80) ? mask : 0 );
    parity ^= d;
    *target = d;
    NM(crc12)(crcBlock, target, 1);
    target++;
    f <<= 1;  /* next bit */
  }  /* for every bit in template from byte */
  if (useParity) {
    *target++ |= parity;
  }

  /* Second byte, odd parity. */
  parity = mask;
  f8 = ft.f800 % 8;
  f = msms[f8];  /* specially-encoded 0.001 seconds */
  for (j=0; j<=3; j++) {
    NM(tD) d;

    d = ( (f & 0x80) ? mask : 0 );
    parity ^= d;
    *target = d;
    NM(crc12)(crcBlock, target, 1);
    target++;
    f <<= 1;  /* next bit */
  }  /* for four MSB bits in template from byte */
  /* Add four MSB bits from CRC-12 generator. */
  *target++ = crcBlock[11];  parity ^= crcBlock[11];
  *target++ = crcBlock[10];  parity ^= crcBlock[10];
  *target++ = crcBlock[9];  parity ^= crcBlock[9];
  *target++ = crcBlock[8];  parity ^= crcBlock[8];
  if (useParity) {
    *target++ = parity;
  }

  /* Third byte, rest of CRC-12 with odd parity. */
  parity = mask;
  for (j=7; j>=0; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }
  return (target);
}  /* mk4fmtf */



NM(ptD)
NM(vlbafmtc)(NM(ptD) target, NM(ptD) crcBlock, int useParity)
{
  unsigned char fts[11];

  /* Init CRC generator to all zeroes. */
  memset(crcBlock, 0, 16*sizeof(NM(tD)));

  /* Emit sync word first as start of a frame. */
  vf.sync[0] = 0xff; vf.sync[1] = 0xff; vf.sync[2] = 0xff; vf.sync[3] = 0xff;
  target = NM(fmt)(target, (ptByte)(&(vf.sync)), 4, -1,
	       useParity?/*even*/-1:0, crcBlock, &(NM(crc16)));

  /* Convert formatter time (MJD, seconds of day) into ASCII-->BCD. */
  /* xxx: mjd => 31-Aug-2132 problem :-) */
  assert( sprintf((char *)fts, "%05d%05d",
		  ft.mjd, ft.dsssss
		  ) <= ((int)sizeof(fts)-1) );
  vf.timec[0] = ((fts[2] & 0x0f) << 4) | (fts[3] & 0x0f);  /* JJ */
  vf.timec[1] = ((fts[4] & 0x0f) << 4) | (fts[5] & 0x0f);  /* JS */
  vf.timec[2] = ((fts[6] & 0x0f) << 4) | (fts[7] & 0x0f);  /* SS */
  vf.timec[3] = ((fts[8] & 0x0f) << 4) | (fts[9] & 0x0f);  /* SS */
  target = NM(fmt)(target, (ptByte)(&(vf.timec)), 4, -1,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc16)));
  /* 'crcBlock' contains "seed" of CRC state this far. */
  return (target);
}  /* vlbafmtc */

NM(ptD)
NM(vlbafmtf)(NM(ptD) target, NM(ptD) crcBlock, int useParity)
{
  tByte f100;
  tByte f8;
  tByte f[2];  /* 0.ssss seconds in BCD */
  NM(tD) parity;
  NM(tD) mask = ~(0);  /* 0xffffffff... */
  int j;

  /* First two bytes, odd parity, .ssss. */
  /* 0.1 and 0.01 seconds, in BCD. */
  f100 = ft.f800 / 8;  /* convert 0.nn seconds to two BCD digits */
  f[0] = (((f100 / 10) & 0x0f) << 4) | ((f100 % 10) & 0x0f) ;
  f8 = ft.f800 % 8;  /* lookup n*1.25msec in lookup table */
  f[1] = vsms[f8];  /* specially-encoded 0.001x seconds */
  target = NM(fmt)(target, f, 2, -1,
	       useParity?/*odd*/1:0, crcBlock, &(NM(crc16)));

  /* 3rd byte, 2xMSnibble of CRC-16, odd parity. */
  parity = mask;
  for (j=15; j>=8; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }

  /* 4th byte, rest of (2xLSnibble of) CRC-16 with odd parity. */
  parity = mask;
  for (j=7; j>=0; j--) {
    *target++ = crcBlock[j];  parity ^= crcBlock[j];
  }
  if (useParity) {
    *target++ = parity;
  }
  return (target);
}  /* vlbafmtf */

NM(ptD)
NM(vlbafmta)(NM(ptD) target, int useParity)
{
  int trk;

  vf.auxsid = 0x12;  /* system ID */
  /* Recorder#, headstack#, hs position. */
  vf.auxhs[0] = 0x00; vf.auxhs[1] = 0x00; vf.auxhs[2] = 0x00;
  target = NM(fmt)(target, (ptByte)(&(vf.auxsid)), 4, -1,
	       useParity?/*odd*/1:0, /*crcBlock=>*/NULL, NULL);

  /* Run track information separately for all bits in data word, */
  /* No need to run last track separately to calc CRC, */
  /* since VLBA aux field is not CRC-protected. */
  for (trk=0; trk<=(BITS-1); trk++) {
    vf.auxdar[0] = trk;  /* BitReversed TG */
    vf.auxdar[1] = trk;  /* BitReversed BS */
    vf.auxdar[2] = trk;  /* BitReversed CD */
    (void)NM(fmt)(target, (ptByte)(&(vf.auxdar)), 3, trk,
	      useParity?/*odd*/1:0, NULL, NULL);
  }  /* for each track */
  target += 3*(useParity?9:8);  /* advance over formatted track data */

  vf.auxunused = 0x00;  /* always zero */
  target = NM(fmt)(target, (ptByte)(&(vf.auxunused)), 1, -1,
	       useParity?/*odd*/1:0, /*crcBlock=>*/NULL, NULL);
  return (target);
}  /* vlbafmta */
