/* wr.c -- Read/write VLBI VSIB input/output board and disks. */

/*
 * $Log: vsib_record.c,v $
 * Revision 1.1  2006/07/19 10:18:47  jwagner
 * added chris' vsib_record files to cvs
 *
 * Revision 1.17.2.10  2006/07/11 04:28:16  phi196
 * Wait until next scan boundary if late starting timed start
 *
 * Revision 1.17.2.9  2006/05/30 09:55:21  phi196
 * Forgot about header size
 *
 * Revision 1.17.2.8  2006/05/30 09:51:43  phi196
 * Adjust filesize for compression modes
 *
 * Revision 1.17.2.7  2006/05/30 09:20:48  phi196
 * Added filename and filesize to network stream
 *
 * Revision 1.17.2.6  2006/03/16 08:01:48  phi196
 * Allow window size to be set
 *
 * Revision 1.17.2.5  2006/03/14 10:54:36  phi196
 * Set header time in net mode
 *
 * Revision 1.21  2006/03/09 02:48:14  phi196
 * Fixed 64 MHz mode
 *
 * Revision 1.20  2006/03/08 13:31:44  phi196
 * Change number of channels based on mode
 *
 * Revision 1.19  2006/03/06 12:35:38  phi196
 * Problem setting time header
 *
 * Revision 1.18  2006/02/10 10:16:41  phi196
 * New header support
 *
 * Revision 1.17.2.4  2006/03/14 07:59:42  phi196
 * Added header support
 *
 * Revision 1.17.2.3  2006/03/12 13:05:38  phi196
 * Larger window
 *
 * Revision 1.17.2.2  2006/02/09 04:46:22  phi196
 * Read remote hostname from command line
 *
 * Revision 1.17.2.1  2006/02/08 22:38:09  phi196
 * Initial TCP networking
 *
 * Revision 1.17  2005/12/23 00:21:33  phi196
 * Minor bug with narrow bands. Check 1 PPS every second. Remove non-used 
 * header stuff. Save filenames to shared memory. Minor clean ups
 *
 * Revision 1.16  2005/10/27 04:21:52  phi196
 * Check all bits for PPS marker for 16 and 32 bit modes
 *
 * Revision 1.15  2005/09/29 01:35:49  phi196
 * Pointer bug
 *
 * Revision 1.14  2005/09/22 06:14:29  phi196
 * Time interval can be float and added (compile time) mode to turn off PSS 
 * check
 *
 * Revision 1.13  2005/09/20 00:11:20  phi196
 * Changed stderr to stdout and force stdout to line buffering (for disko)
 *
 * Revision 1.12  2005/09/12 05:39:16  phi196
 * Convert Zero's to Ohs and 1s to x
 *
 * Revision 1.11  2005/05/09 06:53:22  phi196
 * Clean up some debugging info
 *
 * Revision 1.10  2005/04/14 04:47:47  phi196
 * Added starttime option and autorestart
 *
 * Revision 1.9  2005/03/29 00:22:45  phi196
 * Added starttime option
 *
 * Revision 1.8  2005/03/24 05:30:13  phi196
 * Put compression code in own subroutine
 *
 * Revision 1.7  2005/03/22 04:33:25  phi196
 * Cleaned up code a little ready for auto-restart on 1PPS slip
 *
 * Revision 1.6  2005/03/15 05:42:32  phi196
 * Added hours to -t option
 *
 * Revision 1.5  2005/03/01 09:24:41  phi196
 * Added autofile name(time), default vsib device and bigbuf usage
 *
 * Revision 1.4  2005/02/23 23:29:46  wes128
 * Added <16Mhz support for recording - changed variable rate to clockrate
 *
 * Revision 1.3  2005/01/10 04:15:25  phi196
 * Fixed 1PPS missing bug
 *
 * Revision 1.2  2004/11/10 06:44:03  phi196
 * Forced to write mode
 *
 * Revision 1.1  2004/11/10 03:44:39  phi196
 * Changed name
 *
 * Revision 1.1  2004/11/10 02:53:12  phi196
 * Initial CVS version
 *
 * Revision 1.13  2003/03/05 08:01:56  amn
 * March 2003 Jb2/Mh Mk4/5-style Gbit test version.
 *
 * Revision 1.12  2002/11/29 10:58:40  amn
 * Unified ft2 and ft3; made mrg seq. numbered file opening work; wr parms.
 *
 * Revision 1.11  2002/11/07 09:06:49  amn
 * First Mk5A 1Gbit/s tests with Jodrell.
 *
 * Revision 1.10  2002/11/01 13:19:58  amn
 * Just before changing to use a common tstamp.cc file.
 *
 * Revision 1.9  2002/09/20 07:39:15  amn
 * Added mrg.c, merge/split for multi-PC files.
 *
 * Revision 1.8  2002/09/02 16:39:47  amn
 * Test version suitable for 50MHz/2 test pattern recording, for pb/JB.
 *
 * Revision 1.7  2002/08/09 11:26:56  amn
 * Jul-2002 first fringes Dwingeloo test version.
 *
 * Revision 1.5  2002/06/15 10:07:31  amn
 * Added shared memory and seek support.
 *
 * Revision 1.4  2002/06/14 13:02:34  amn
 * Dwingeloo test trip version.
 *
 */

/* Copyright (C) 2001--2002 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <errno.h>

#include <unistd.h>
#include <time.h>
#include <math.h> /* drem() */

#include <sys/types.h>  /* open() */
#include <sys/stat.h>  /* open() */
#include <fcntl.h>  /* open() */

#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>  /* gettimeofday(), usleep(), getopt() */

#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

extern char *optarg;
extern int optind, opterr, optopt;

#include <string.h>  /* strstr(),strtok() */

#include <sys/ipc.h>  /* for shared memory */
#include <sys/shm.h>

#include <sys/ioctl.h>
#include "vsib_ioctl.h"

#include "chris/tstamp.h"
#include "chris/vheader.h"
#include "chris/sh.h"


#define STRLEN 256

#define FALSE 0
#define TRUE 1

#define AUTO_RESTART

#define MINSTART_USEC 300000
#define MAXSTART_USEC 800000

#define TIMESTAMP(m) {fprintf(stderr, "**** %s at %.4f\n", m, tim());}

/* VSIB recording modes */

/* Depending on readMode either standard input or output. */
int vsib_fileno;

void start_recorder (int mode, int skip, char datestr[], char timestr[],
		     struct tm **date, double *starttime);
void stop_recorder (int *usleeps);
void waituntil (double starttime, float fileint);
time_t toUT (time_t time);
void compress_data (char *mem, int c_mode, size_t nread, size_t *nwrite);
int setup_net(char *hostname, int port, int window_size, int *sock);
int netsend(int sock, char *buf, size_t len);

#define TIMESTART(i) \
        { dirs[i].prevOperTime = tim(); }
#define TIMESTOP(i) \
      { double dur = tim() - dirs[i].prevOperTime; \
        if (dur > dirs[i].maxOperTime) dirs[i].maxOperTime = dur; }

/* A protected (error-checking) ioctl() for VSIB driver. */
static void
vsib_ioctl(
  unsigned int mode,
  unsigned long arg
) {
  if (ioctl(vsib_fileno,
            mode,
            arg)
  ) {
    char *which;
    char err[255];
    
    if (vsib_fileno == STDOUT_FILENO) {
      which = "rd";
    } else {
      which = "wr";
    }
    snprintf(err, sizeof(err), "%s: ioctl(vsib_fileno, 0x%04x,...)", which, 
	     mode);
    perror(err);
    fprintf(stderr, "%s: standard I/O is not an VSIB board\n", which);
    exit(EXIT_FAILURE);
  }
}  /* vsib_ioctl */


typedef struct sDir {
  char *nameTemplate;
  int nameNumber;
  int blocksPerFile;
  int fid;
  int nowBlocks;
  double prevOperTime;
  double maxOperTime;
} tDir, *pDir;


double tim(void) {
  struct timeval tv;
  double t;

  assert( gettimeofday(&tv, NULL) == 0 );
  t = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;
  return t;
}  /* tim */

void
start_VSIB(int mode, int skip) {
  vsib_ioctl(VSIB_SET_MODE,
	     (VSIB_MODE_MODE(mode)
	      | VSIB_MODE_RUN
	      | (0 ? VSIB_MODE_GIGABIT : 0)
	      | VSIB_MODE_EMBED_1PPS_MARKERS
	      | (skip & 0x0000ffff))
	     );
}  /* start_VSIB */


int relSeekBlocks; 

/* Global Variables for the time being. These should go */

int verbose = TRUE;

int
main(int argc, char *argv[]) {
  // char *wr_version="1.0.1"; // Added 1PPS check
  // char *wr_version="1.0.2"; // Added Cmode to the header
  // char *wr_version="1.0.3"; // Added a bigbuffer check
  // char *wr_version="1.0.4"; // Increase the effect of "verbose" mode
  // char *wr_version="1.0.5"; // Cleaned up a little and added headers as 
                               // an option
  // char *wr_version="1.0.6"; // Bandwidth ajustable to less than 16 MHz - cwest
  // char *wr_version="1.0.7";   // Added starttime option and autorestart - cjp
  // char *wr_version="1.8";     // Various stuff
  char *wr_version="1.9";      // Added extended headers

  int readMode;
  int blksize=32000;
  int totalblks=3600; /* 1 hour */
  int memblks;
  char *memblk,tmp_ptr;
  char *p,*fileout=NULL;
  char *vsib_device=NULL;
  int ndirs=1;
  pDir dirs;
  int i;
  double recordStart;
  double totalDur;
  double startTime=0;
  double elapsed;
  double gmt;
  int usleeps;
  int bytespersample=0;
  int blocksPerSec;
  int vsib_started = 0;
  int blk=10; /* 10 seconds */
  float fileint = 0.0; /* Remember file size in seconds */
  double mbits;
  int clockrate=32; // Default rate 32 MHz
  int bandwidth=16; // Default bandwidth 16MHz - for values less than 16Mhz 
                    // sample skipping is needed.
  int skip=0;       // Used for bandwidth < 16MHz
  int nchan=4;      // Number of channels recorded

  char *cptr;
  int ppsOK;
  int loop=FALSE,tb_in_sec=TRUE,fb_in_sec=TRUE;
  int c_mode=15,mode=3;
  int cfact = 1;
  int net=0;
  int port=8000;
  int window_size = 2000 * 1024;	/* 2 MB */
  int sock;
  int status;
  int scale;
  int newfile=FALSE;
  char hostname[STRLEN] = ""; /* Host name to send data to */
  char *headbuf;
  vhead *header;    // File header object

  /* Shared memory unique identifier (key) and the returned reference id. */
  key_t shKey;
  int shId = -1;
  ptSh sh;

  char datestr[100], timestr[100], msg[100];
  struct timeval current;
  struct timezone zone;
  struct tm *date;
  time_t now;

  while ((tmp_ptr = getopt(argc, argv, "H:s:e:w:W:r:t:b:f:m:c:d:p:i:o:lvh")) != EOF)
    switch (tmp_ptr) {
    case 'w': // recording bandwidth
      bandwidth=atoi(optarg);
      if(verbose) printf("Bandwidth: %d MHz\n", bandwidth);
      break;

    case 'W': // Window size
      window_size=atoi(optarg)*1024;
      if(verbose) printf("Windowsize: %d kBytes\n", window_size);
      break;

    case 'r': // recording rate
      clockrate=atoi(optarg);
      if (verbose) printf("Rate: %d\n",clockrate);
      if (!clockrate) exit(printf("Rate given (%s) is zero\n",optarg));
      break;

    case 'b': // block size
      blksize=atoi(optarg);
      if (verbose) printf("Blocksize: %d\n",blksize);
      if (!blksize) 
	exit(printf("Block size given (%s) is zero\n",optarg));
      break;

    case 'f': // blocks per file
      if (optarg[strlen(optarg)-1]=='s') 
	fb_in_sec=TRUE;
      else  
	fb_in_sec=FALSE;
      blk=atoi(optarg);
      if (verbose) printf("Blocks per file: %s\n",optarg);
      if (!blk) 
	exit(printf("Blocks per file given (%s) is zero\n",optarg));
      break;

    case 't': // seconds to record
      scale=1;
      if (optarg[strlen(optarg)-1]=='s') 
	tb_in_sec=TRUE;
      else if (optarg[strlen(optarg)-1]=='h') {
        scale=60*60; /* Hours -> seconds  */
	tb_in_sec=TRUE;
      } else if (optarg[strlen(optarg)-1]=='m') {
        scale=60; /* Minutes -> seconds  */
	tb_in_sec=TRUE;
      } else  
	tb_in_sec=FALSE;

      totalblks=ceil(atof(optarg)*scale);
      if (verbose) fprintf(stderr,"Recording time: %d\n",totalblks);
      if (verbose) printf("Recording time: %d\n",totalblks);

      if (!totalblks) 
	exit(printf("Time to record (%s) is zero\n",optarg));
      break;

    case 'd': // Number of directories
      ndirs=atoi(optarg);
      if (verbose) printf("Number of directories: %d\n",ndirs);
      if (!ndirs) 
	exit(printf("Number of directories to record (%s) is zero\n",
		     optarg));
      break;

    case 'c': // Compress on the fly
      // First convert Zeros to Ohs and 1s to x
      for (cptr=optarg; *cptr!=(char)0; cptr++) {
	if (*cptr=='0')
	  *cptr = 'o';
	else if (*cptr=='O')
	  *cptr = 'o';
	else if (*cptr=='1')
	  *cptr = 'x';
      }

      if (!strcmp(optarg,"oooo")) {
	c_mode=0; nchan=0; printf("No data!\n");

      } else if (!strcmp(optarg,"ooox")) { 
	c_mode=1; cfact=4; nchan=1; if (verbose) printf("Channel one only\n");

      } else if (!strcmp(optarg,"ooxo")) { 
	c_mode=2; cfact=4; nchan=1; if (verbose) printf("Channel two only\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"ooxx")) { 
	c_mode=3; cfact=2; nchan=2; if (verbose) printf("Channel one & two\n");

      } else if (!strcmp(optarg,"oxoo")) { 
	c_mode=4; nchan=1; if (verbose) printf("Channel three only\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"oxox")) { 
	c_mode=5; cfact=2; nchan=2; if (verbose) printf("Channel one & three\n");

      } else if (!strcmp(optarg,"oxxo")) { 
	c_mode=6; cfact=2; nchan=2; if (verbose) printf("Channel two & three\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"oxxx")) { 
	c_mode=7; cfact=0; nchan=3; if (verbose) printf("Channel one, two & three\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xooo")) { 
	c_mode=8; cfact=4; nchan=1; if (verbose) printf("Channel four only\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xoox")) { 
	c_mode=9; cfact=2; nchan=2; if (verbose) printf("Channel one & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xoxo")) { 
	c_mode=10; cfact=2; nchan=2;  if (verbose) printf("Channel two & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xoxx")) { 
	c_mode=11; cfact=-1; nchan=3; if (verbose) printf("Channel one, two & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xxoo")) { 
	c_mode=12; cfact=2; nchan=2; if (verbose) printf("Channel three & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xxox")) { 
	c_mode=13; cfact=-1; nchan=3; if (verbose) printf("Channel one, three & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xxxo")) { 
	c_mode=14; cfact=-1; nchan=3; if (verbose) printf("Channel two, three & four\n");
	exit(printf("Not yet supported!\n"));

      } else if (!strcmp(optarg,"xxxx")) { 
	c_mode=15; cfact=1; nchan=4; if (verbose) printf("No compression\n");

      } else { 
	exit(printf("Unknown channel mode %s\n",optarg)); 
      }
      break; 

    case 'o': // file to record
      fileout=(char *) strdup(optarg);
      if (verbose) printf("Disk File: %s\n",fileout);
      if (!fileout) 
	exit(printf("File to record to (%s) is zero\n",optarg));
      break;

    case 'H': // Remote host to connect to
      if (strlen(optarg)>STRLEN-1) {
	fprintf(stderr, "Warning: Remote host (%s) too long\n", optarg);
	exit(1);
      } 
      strcpy(hostname, optarg);
      net = 1;
      if (verbose) printf("Remote host: %s\n",hostname);
      break;

    case 'e': // vsib device file
      vsib_device=(char *) strdup(optarg);
      if (verbose) printf("Vsib device: %s\n",vsib_device);
      if (!vsib_device) 
	exit(printf("Error setting vsib device to \"%s\"\n",optarg));
      break;

    case 'm': // mode to record
      mode=atoi(optarg);
      if (verbose) printf("Mode: %d\n",mode);
      if (mode<0 || mode>13) 
	exit(printf("invalid mode %s\n",optarg));
      if (mode>4) 
	printf("invalid mode %s\n",optarg);
      break;

    case 's': // start time
      /* Get and convert target UTC time to Unix seconds. */
      if (getDateTime(optarg, &startTime)) {
	printf("Failed to convert ISO 8601 UTC date/time `%s'\n", 
		optarg);
	exit(EXIT_FAILURE);
      }
      break;

    case 'l': // loop recording
      loop=TRUE;
      if (verbose) printf("Looping mode\n");
      break;

    case 'p':
      port = atoi(optarg);
      if (verbose) printf("Port: %d\n", port);
      net = 1;
      break;

    case 'v': // verbose
      verbose=TRUE;
      if (verbose) printf("Verbose mode\n");
      break;

    case 'h': // help or unknown
    case '?':
      exit(printf(
	 "usage : %s -r rate(MHz) -t recording_time(blks/secs/hours) "
		   "-b blksize -f blk -o filenameprefix -c compact_mode "
		   "-m mode -l (loop) -h (help) -v (verbose) "
		   "-w bandwidth(MHz)\nVersion %s\n", argv[0],
		   wr_version));
    }
  
  if (tb_in_sec || fb_in_sec) {
    switch (mode) {
    case 0:
      bytespersample = 4;
      break;
    case 2:
      bytespersample = 2;
      break;
    case 3:
      bytespersample = 1;
      break;
    default:
      printf("Warning: Unsupported recording mode %d. "
	      "Time calculations\n will probably be wrong\n", mode);
      bytespersample = 1;
      break;
    }
  }

  if(bandwidth*2 < clockrate){
    skip = (clockrate/bandwidth/2) - 1;
  }

  if (tb_in_sec) totalblks=(((float)totalblks)*clockrate*1000000
			    *bytespersample)/blksize/(skip+1);
  if (fb_in_sec) {
    fileint = blk;
    blk=(((float)blk)*clockrate*1000000*bytespersample)
      /blksize/(skip+1);
  } else
    fileint = blk*blksize*(skip+1)/(clockrate*1e6*bytespersample);

  blocksPerSec = clockrate*1e6*bytespersample/blksize/(skip+1);


  if (verbose) printf("Total number of blocks: %d\n",totalblks);
  if (verbose) printf("blocksize: %d\n",blksize);
  if (verbose) printf("No. blocks per file: %d\n",blk);
  if (!fileout) fileout=strdup("test");
  if (!vsib_device) vsib_device=strdup("/dev/vsib");

  /* Init variables according to command line, allocate buffers. */
  /*readMode = (strstr(argv[0], "rd") != NULL);*/
  readMode = 0;

  if (readMode) {
    vsib_fileno = open(vsib_device, O_WRONLY);
  } else {
    vsib_fileno = open(vsib_device, O_RDONLY);
  }
  if (vsib_fileno == -1) {
    snprintf(msg, 100, "Trying to open %s", vsib_device);
    perror(msg);
    exit(0);
  }

  memblks = 1;
  assert( (memblk = malloc(memblks*blksize)) != NULL );

  /* Allocate one extra struct for call-to-call time statistics. */
  assert( (dirs = malloc((ndirs+1)*sizeof(tDir))) != NULL );
  for (i=0; i<ndirs; i++) {
    if (!i) dirs[i].nameTemplate = strtok(fileout," ");
    else dirs[i].nameTemplate = strtok(NULL," ");
    if (!dirs[i].nameTemplate) 
      exit(printf("name template %d failed\n",i));
    dirs[i].nameNumber = 0;
    dirs[i].blocksPerFile = blk;
    dirs[i].nowBlocks = dirs[i].blocksPerFile;  /* init to end */
    dirs[i].maxOperTime = 0.0;  /* init to smallest value */
  }  /* for each directory */
  dirs[ndirs].nameTemplate =   "call-to-call  ";
  dirs[ndirs].maxOperTime = 0.0;

  /* Create and initialize shared memory. */
  shKey = fourCharLong('v','s','i','b');
  assert( (shId = shmget(shKey, sizeof(tSh), IPC_CREAT | 0777)) != -1 );
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (void *)-1 );
  relSeekBlocks = 0;

  /* Initilize shared memory values */
  strcpy(sh->currentfile, ""); 

  /* Initialize the header object and fill in the static values */

  header = newheader();

  if (mode==2) 
    nchan *= 2;
  else if (mode==0)
    nchan *= 4;

  if (bandwidth==64) 
    nchan /= 4;
  else if (bandwidth==32) 
    nchan /= 2;

  setantennaid(header, "Tt");
  setantennaname(header, "Test Antenna");
  setexperimentid(header, "VTXXXX");
  setrecorderversion(header, 1, 0);
  setbandwidth(header, bandwidth);
  setnchan(header, nchan);
  setnumbits(header, 2);
  setencoding(header, AT);
 
  /* If requested wait until the specified time */

  if (startTime!=0.0) {
    waituntil(startTime-0.5, fileint);
  }

#if 0
  Opening the device special file already does this.
  For write() this open()-related (extra) DMA reset clears
  memory buffers and r/w pointers with arg==0.
  /* Init DMA. */
  vsib_ioctl(VSIB_RESET_DMA, 0);
#endif

  setlinebuf(stdout);

  if (net) {
    /* Connect to the receiving server */

    if (strlen(hostname) == 0) {
      strcpy(hostname, "localhost");
    }

    status = setup_net(hostname, port, window_size, &sock);
    if (status) exit(1);
  }

  /* Start the board.  (After opening it is still stopped.) */
  /* For disk write, start VSIB first and write to disk later. */
  /* For disk read, read first and start VSIB later. */
  if (readMode) {
    vsib_started = 0;
  } else {
    start_recorder(mode, skip, datestr, timestr, &date, &startTime);
  }

  /* Start looping for total number of blocks. */
  usleeps = 0;
  recordStart = tim();
  TIMESTART(ndirs);  /* call-to-call (not per file) */

  blk=0;
  while (blk<totalblks) {
    i = blk % ndirs;
    if (relSeekBlocks) {
      int bl;

      /* Adjust 'blk' (and i and newBlocks) based on 'relSeekBlocks'. */
      /* xxx: initially only within a single open file */
      bl = dirs[i].nowBlocks + relSeekBlocks;
      if (bl < 0) {
	relSeekBlocks -= bl;
	bl = 0;
      }
      bl = dirs[i].nowBlocks + relSeekBlocks;
      if (bl > dirs[i].blocksPerFile) {
	relSeekBlocks -= (bl - dirs[i].blocksPerFile);
	bl = dirs[i].blocksPerFile;
      }
      assert( lseek(dirs[i].fid, (bl*blksize), SEEK_SET) != (off_t)-1 );
      printf(
	      "at block = %d, repositioning %+d blocks (to byte offset %d)\n",
	      blk,
	      relSeekBlocks,
	      (bl*blksize)
	      );
      dirs[i].nowBlocks = bl;
      blk += relSeekBlocks;
      /* xxx: cannot really update i... */

      /* Seeking done, notify shared memory. */
      relSeekBlocks = 0;
    }  /* if external adjustments */

    if (dirs[i].nowBlocks >= dirs[i].blocksPerFile) {
      /* Open a new file in this directory / template. */
      char pn[255];
      int n = blk / dirs[i].blocksPerFile;

      snprintf(pn, sizeof(pn), "%s_%s.lba", dirs[i].nameTemplate, timestr);

      if (n==0) TIMESTART(i);  /* only for first file */

      newfile=TRUE;

      setheadertime(header, date);
      if (net) {
	unsigned long long filesize;
	short fnamesize;

	writeheader(header, 0, &headbuf);

	/* Send network header */
	filesize = dirs[i].blocksPerFile * blksize / cfact + header->headersize;
	fnamesize = strlen(pn)+1;
	status = netsend(sock, (char*)&filesize, sizeof(long long));
	if (status) exit(1);
	status = netsend(sock, (char*)&fnamesize, sizeof(short));
	if (status) exit(1);
	status = netsend(sock, pn, fnamesize);
	if (status) exit(1);

	status = netsend(sock, headbuf, header->headersize);
	if (status) exit(1);
	free(headbuf);
      } else { // Writing to disk
	if (readMode) {
	  assert((dirs[i].fid=open(pn, O_RDONLY)) != -1 );
	} else {
	  assert((dirs[i].fid=creat(pn, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))!=-1);

	  // Write the header into the data file
	  
	  writeheader(header, dirs[i].fid, NULL);

	}
      } 
      system("date");

      printf("at block = %d, opened file '%s'\n", blk, pn);
      dirs[i].nowBlocks = 0;

      strcpy(sh->lastfile, sh->currentfile); // Remember last filename
      if (strlen(pn)>=SH_MAXSTR) {
	printf("%s is too long to save to shared memory, increase MAXSTR in sh.h\n", pn);
	strcpy(sh->currentfile, ""); 
      } else 
	strcpy(sh->currentfile, pn); // Save filename to shared memory. 
      
    }

    /* Write (or read) one block. */
    /* xxx: need to add '-1' error checks to VSIB read()/write() calls */
    p = &(memblk[(blk % memblks)*blksize]);
    if (readMode) {
      size_t nwritten;

      assert( (read(dirs[i].fid,
                     p,
                     blksize)) == blksize );

      /* Write a block to VSIB; if not enough space, sleep a little. */
      nwritten = write (vsib_fileno, memblk, blksize);
      while (nwritten < (size_t) blksize) {
        /* Not space for one full block in buffer, start VSIB transmission */
        /* and wait for vacant space. */
        if (readMode) {
          if (!vsib_started) {
            printf("starting vsib because bigbuf full\n");
            /* xxx: reload PLX FIFOs */
            vsib_ioctl(VSIB_RESET_DMA, /*skip_clear=>*/1);
            start_VSIB(mode, skip);
            vsib_started = 1;
          }
        }
        usleep(1000);  /* a small amount, probably ends up to be 10--20msec */
        usleeps++;
        nwritten += write (vsib_fileno, memblk+nwritten, blksize-nwritten);
      }  /* while not at least one full block in VSIB DMA ring buffer */

    } else {
      size_t nread,nwrite;

      /* Read a block from VSIB; if not enough, sleep a little. */
      nread = read (vsib_fileno, memblk, blksize);
      while (nread < (size_t) blksize) {
        /* Not one full block in buffer, wait. */
        usleep(1000);  /* a small amount, probably ends up to be 10--20msec */
        usleeps++;
        nread += read (vsib_fileno, memblk+nread, blksize-nread);
      }  /* while not at least one full block in VSIB DMA ring buffer */
      if (newfile) {
	newfile=FALSE;

      }

      if (!dirs[i].nowBlocks) {  // new file

	if (verbose) { 
	  unsigned long b, bigbuf_size;

	  vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&b);

	  /* Find out total size of bigbuf area */
	  vsib_ioctl(VSIB_GET_BIGBUF_SIZE, (unsigned long)&bigbuf_size);

	  printf("MBytes left in the BIGBUF %lu (%2.0f%%): ", 
		  b/1024/1024, b/(double)bigbuf_size*100); 
	}
      }


      if (!(dirs[i].nowBlocks % blocksPerSec)) {  // 1 sec boundary

	// Check PPS marker is present. Need to handle 1, 2 and 4 byte data
	ppsOK = 0;
	if ((*memblk&0xFF)==0xFF) {
	  if (mode==2 || mode==0) {
	    char *memptr;
	    memptr = memblk;
	    memptr++;

	    if ((*memptr&0xFF)==0xFF) {
	      if (mode==0) {
		memptr++;
		if ((*memptr&0xFF)==0xFF) {
		  memptr++;
		  if ((*memptr&0xFF)==0xFF) {
		    ppsOK = 1;
		  }
		}
	      } else {
		ppsOK = 1;
	      }
	    }
	  } else {
	    ppsOK = 1;
	  }
        }

	if (ppsOK) {
	  if (verbose && dirs[i].nowBlocks==0) printf("1PPS OKAY\n");
	}
#ifdef AUTO_RESTART
        else {
	  double wait;

	  printf("\n\nERROR: 1PPS transition ABSENT (%02x) at %s\n",
		  *memblk,datestr);
	  printf(" Restarting recorder\n");

	  // First stop we must stop the recorder
	  stop_recorder(&usleeps);

	  // Need to re-initialise file struct to ensure files are re-opened
	  for (i=0; i<ndirs; i++) {
	    dirs[i].nowBlocks = dirs[i].blocksPerFile;  /* init to end */
	  }

	  // Close and re-open device file
	  if (close(vsib_fileno)) {
	    perror("Trying to close vsib device file");
	    exit(0);
	  }
	  if (readMode) {
	    vsib_fileno = open(vsib_device, O_WRONLY);
	  } else {
	    vsib_fileno = open(vsib_device, O_RDONLY);
	  }
	  if (vsib_fileno == -1) {
	    snprintf(msg, 100, "Trying to re-open %s", vsib_device);
	    perror(msg);
	    exit(0);
	  }

	  // Wait so we start on the same integration time boundary as
	  // we previously were


	  gettimeofday(&current, NULL);
	  if (current.tv_usec > 500000) current.tv_sec++; // (0.5 sec) Avoid timing issues

	  gmt = toUT(current.tv_sec + 1); // This is the time we want to start at

	  elapsed = toUT(current.tv_sec + 1) - startTime;

	  printf("*** Have been recording for %.1f sec\n", elapsed);

	  wait = (fileint-fmod(elapsed, fileint));

	  printf("*** Will wait for %.1f sec\n", wait);

	  gmt += wait; // Need to add this delay to start at the right place
	  waituntil(gmt-0.5, fileint);

	  // Now restart the recorder
	  start_recorder(mode, skip, datestr, timestr, &date, NULL);
	  continue; // Restart loop
	}
#endif  //AUTO_RESTART
      }
    
      nwrite=nread;  // Write it all out

      if (c_mode==0) {
	nwrite=0; // Skip the write
      } else if (c_mode != 15) 
	compress_data(memblk, c_mode, nread, &nwrite);

      if (net) {
	status = netsend(sock, p, nwrite);
	if (status) exit(1);
      } else {
	assert( (write(dirs[i].fid, p, nwrite)) == (ssize_t)nwrite );
      }
    }
    TIMESTOP(i);  /* accumulates close--open-write in worst case */
    TIMESTOP(ndirs);  /* call-to-call */

    TIMESTART(i);
    TIMESTART(ndirs);
    dirs[i].nowBlocks++;

    /* Time to switch to another file? */
    if (dirs[i].nowBlocks >= dirs[i].blocksPerFile) {
      /* Close the ready file before opening a new. */
      if (!net) assert( (close(dirs[i].fid) != -1) );

      gettimeofday(&current, &zone);

      if(current.tv_usec/1000000.0 > 0.5)
	now = current.tv_sec + 1;
      else
	now = current.tv_sec;

      date = gmtime(&now); // will start on the next second
      strftime(datestr, 99, "%Y%m%d:%H%M%S\n", date);
      strftime(timestr, 99, "%j_%H%M%S", date);
    }
		blk++;
    if (loop) blk=blk%totalblks;
  }  /* while loop */
  TIMESTOP(ndirs);
  totalDur = tim() - recordStart;

  /* Wait for the data to get out of VSIB. */
  if (readMode) {
    unsigned long b;

    if (!vsib_started) {
      printf("all written to bigbuf, starting vsib...\n");
      /* xxx: reload PLX FIFOs */
      vsib_ioctl(VSIB_RESET_DMA, /*skip_clear=>*/1);
      start_VSIB(mode, skip);
      usleep(2000000);  /* xxx */
    }

    vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&b);
    while (b > 0) {
      printf("Waiting for ring buffer to empty (%lu bytes, sl=%d)\n",
              b, usleeps);
      /* Still bytes in bigbuf waiting for transmit. */
      usleep(500000);  /* 0.5sec */
      usleeps++;
      vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&b);
    }  /* while not bigbuf empty */

  }  /* if readMode, wait for data to get out of ring buffer */

  stop_recorder(&usleeps);

  /* Final report. */
  mbits = 8.0*totalblks*blksize/totalDur/1000.0/1000.0;
  if (verbose) printf(
         "Took %f seconds, %f Mbits(dec)/s (%.1f%% of PCI33), usleeps = %d.\n",
         totalDur,
         mbits,
         mbits/(33.0*4.0*8.0)*100.0,
         usleeps
         );
  for (i=0; i<=ndirs; i++) {
    printf("%s  max. %f seconds.\n",
           dirs[i].nameTemplate,
           dirs[i].maxOperTime
           );
  }  /* for each directory */

  /* Remove shared memory to mark that 'wr/rd' is no more running. */
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was allocated

  return(EXIT_SUCCESS);
}  /* main */


void start_recorder (int mode, int skip, char datestr[], char timestr[],
		     struct tm **date, double *starttime) {
  struct timeval current;
  struct timezone zone;
  time_t now;

  // Delay the record start to a range of (0.3 -> 0.8) seconds
  // This ensures the program knows what second it is starting on.

  gettimeofday(&current, &zone);
  if (verbose) printf("%0.2f\n", current.tv_usec/1000000.0);
  while(current.tv_usec < MINSTART_USEC || current.tv_usec > MAXSTART_USEC){
    usleep(100);
    gettimeofday(&current, &zone);

  }

  now = current.tv_sec + 1;

  if (starttime != NULL) {
    *starttime = (double)toUT(now);
  }

  *date = gmtime(&now); // will start on the next second
  strftime(datestr, 99, "%Y %B %e %H:%M:%S UTC", *date);
  fprintf(stdout, "Start will occur at: %s\n", datestr);
  strftime(datestr, 99, "%Y%m%d:%H%M%S\n", *date);
  strftime(timestr, 99, "%j_%H%M%S", *date);

  start_VSIB(mode, skip);
}


/* The vsib device file seems to have to be closed before it can be used
   again. This could be added to this routine */

void stop_recorder (int *usleeps) {
  unsigned long b;

  /* Stop the board, first DMA, and when the last descriptor */
  /* has been transferred, then write stop to board command register. */

  vsib_ioctl(VSIB_DELAYED_STOP_DMA, 0);

  vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
  while (!b) {
    printf("Waiting for last DMA descriptor (sl=%d)\n",
	    *usleeps);
    usleep(100000);
    (*usleeps)++;
    vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
  }

  vsib_ioctl(VSIB_SET_MODE, VSIB_MODE_STOP);
}

time_t toUT (time_t time) {
  struct tm *date;

  date = gmtime(&time);
  return mktime(date);
}

 void waituntil (double starttime, float fileint) {
  double now;
  time_t nowsec, nowgmt;

  now = tim();

  /* We cannot be sure what timezone gettimeofday will return in
     We want GMT. The following should give us the offset in seconds */ 

  nowsec = now;
  nowgmt = toUT(nowsec);
  starttime += nowsec-nowgmt;

  if (now > starttime) {
    printf("Clock already past start time `%s'\n", "????");
    
    while (starttime<now) starttime += fileint; // Advance to the next boundary
  } 

  while (now < starttime) {
    double stillLeft = starttime - now;
    double newSleep;
    unsigned long sleepusecs;

    /* Sleep half of the available time. */
    newSleep = stillLeft / 2;
    sleepusecs = (unsigned long)(newSleep * 1000000.0);

    printf("%.2f seconds until start, sleeping %.2f seconds...\n", 
	    stillLeft, newSleep);
    usleep(sleepusecs);

    now = tim();
  }  /* while time left */

  return;
}

void compress_data (char *mem, int c_mode, size_t nread, size_t *nwrite) {
  char *iptr,*optr;
  int j;

  iptr=optr=mem;

  switch (c_mode) {
  case 1:  // Channel 1
    for (j=0; j<(int) nread/4; j++) {
      *optr =  *iptr & 0x03;
      iptr++;
      *optr |= (*iptr & 0x03)<<2;
      iptr++;
      *optr |= (*iptr & 0x03)<<4;
      iptr++;
      *optr |= (*iptr & 0x03)<<6;
      iptr++;
      optr++;
    }
    *nwrite = nread/4;
    break; 

  case 3:   // Channel 1+2
    for (j=0; j<(int) nread/2; j++) {
      *optr =  *iptr & 0x0F;
      iptr++;
      *optr |= (*iptr & 0x0F)<<4;
      iptr++;
      optr++;
    }
    *nwrite = nread/2;
    break;
  case 5:   // Channel 1+3
    for (j=0; j<(int) nread/2; j++) {
      *optr = (*iptr & 0x03) | (*iptr & 0x30)>>2;
      iptr++;
      *optr |= (*iptr & 0x03)<<4 | (*iptr & 0x30)<<2;
      iptr++;
      optr++;
    }
    *nwrite = nread/2;
    break;
  default:
    printf("Warning: Do not support compressing mode %d\n", c_mode);
  }
}

void randbuf(char *buf, ssize_t nbytes) {
  int i;

  for (i=0; i<nbytes; i++) {
    buf[i] = (char)rand();
  }  
}  

int setup_net(char *hostname, int port, int window_size, int *sock) {
  int status;
  unsigned long ip_addr;
  struct hostent     *hostptr;
  struct linger      linger = {1, 1};
  struct sockaddr_in server;    /* Socket address */

  hostptr = gethostbyname(hostname);
  if (hostptr==NULL) {
    fprintf(stderr,"Failed to look up hostname %s\n", hostname);
    return(1);
  }
  
  memcpy(&ip_addr, (char *)hostptr->h_addr, sizeof(ip_addr));
  memset((char *) &server, 0, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_port = htons((unsigned short)port); 
  server.sin_addr.s_addr = ip_addr;
  
  printf("Connecting to %s\n",inet_ntoa(server.sin_addr));
    
  *sock = socket(AF_INET, SOCK_STREAM, 0);
  if (*sock==-1) {
    perror("Failed to allocate socket");
    return(1);
  }

  /* Set the linger option so that if we need to send a message and
     close the socket, the message shouldn't get lost */
  status = setsockopt(*sock, SOL_SOCKET, SO_LINGER, (char *)&linger,
		      sizeof(struct linger)); 
  if (status!=0) {
    close(*sock);
    perror("Setting socket options");
    return(1);
  }

  /* Set the window size to TCP actually works */
  status = setsockopt(*sock, SOL_SOCKET, SO_SNDBUF,
		      (char *) &window_size, sizeof(window_size));
  if (status!=0) {
    close(*sock);
    perror("Setting socket options");
    return(1);
  }
  status = setsockopt(*sock, SOL_SOCKET, SO_RCVBUF,
		      (char *) &window_size, sizeof(window_size));
  if (status!=0) {
    close(*sock);
    perror("Setting socket options");
    return(1);
  }
    
  status = connect(*sock, (struct sockaddr *) &server, sizeof(server));
  if (status!=0) {
    perror("Failed to connect to server");
    return(1);
  }
  return(0);
} /* Setup Net */


int netsend(int sock, char *buf, size_t len) {
  char *ptr;
  int ntowrite, nwrote;
  
  ptr = buf;
  ntowrite = len;

  while (ntowrite>0) {
    nwrote = send(sock, ptr, ntowrite, 0);
    if (nwrote==-1) {
      if (errno == EINTR) continue;
      perror("Error writing to network");

      return(1);
    } else if (nwrote==0) {
      fprintf(stderr, "Warning: Did not write any bytes!\n");
      return(1);
    } else {
      ntowrite -= nwrote;
      ptr += nwrote;
    }
  }
  return(0);
}
