/*
 * Tests the fseek() and fread() functions of the VSIB driver
 *
 * Reads one large block of data from the VSIB, then reads
 * it again in smaller blocks and confirms that the data matches.
 *
 * (C) 2007 Jan Wagner
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>

#include "vsib_ioctl.h"
#define BS_DEFAULT 32768
#define MAX_ITER   40
#define INNER_ITER 10

int readvsib(FILE *fp, unsigned char *b, size_t count) {
   int read_iter = 0;
   size_t rd = 0, ret;
   do {
      ret = fread(b + rd, 1, count - rd, fp);
      rd += ret;
      read_iter++;
   } while(rd<count);
   return read_iter;
}

int main(int argc, char *argv[])
{
   int    blocksize = 0;
   int    read_iter = 0;
   int    failed = 0;
   off_t  offset = 0L;
   FILE  *fpVSIB = NULL;
   unsigned char *buf;
   unsigned char *sbuf[INNER_ITER];
   unsigned char *tmpbuf;
   int i, j;

   if (argc > 1) {
       blocksize = atoi(argv[1]);
   }
   if (blocksize < 128) {
       blocksize = BS_DEFAULT;
   }
   buf = malloc(INNER_ITER*blocksize);
   tmpbuf = malloc(blocksize);
   for (i=0; i<INNER_ITER; i++) {
       sbuf[i] = &buf[i*blocksize];
   }
   fprintf(stderr, "Using block size: %d\n", blocksize);

   fpVSIB  = fopen64("/dev/vsib", "r");
   ioctl(fileno(fpVSIB), VSIB_SET_MODE, (VSIB_MODE_MODE(0) | VSIB_MODE_RUN | 4));

   for (i=0; i<MAX_ITER; i++) {
      // if ((i%50)==0) fprintf(stderr, "Iteration %d...\n", i);
      fprintf(stderr, "Iter %05d/%d : ", i, MAX_ITER);
      fseeko64(fpVSIB, ((off_t)blocksize) * INNER_ITER * i, SEEK_SET);
      readvsib(fpVSIB, buf, INNER_ITER*blocksize);
      read_iter = 0;
      for (j=0; j<INNER_ITER; j++) {
         memcpy(tmpbuf, sbuf[j], blocksize);
         fseeko64(fpVSIB, ((off_t)blocksize) * (INNER_ITER*i + j), SEEK_SET);
         read_iter += readvsib(fpVSIB, sbuf[j], blocksize);
         if (0 != memcmp(tmpbuf, sbuf[j], blocksize)) {
            // fprintf(stderr, "mismatch at iter %d:%d\n", i, j);
            fprintf(stderr, "1");
            failed++;
         } else {
            fprintf(stderr, "0");
         }
      }
      fprintf(stderr, " (%d read()s)\n", read_iter);
   } 

   ioctl(fileno(fpVSIB), VSIB_SET_MODE, VSIB_MODE_STOP);
   fclose(fpVSIB);

   fprintf(stderr, "Total blocks (re)read : %d\n", (1 + INNER_ITER) * MAX_ITER);
   fprintf(stderr, "Total failures count  : %d\nDone.\n", failed);
}
